import sys
import os

from io import TextIOBase
from argparse import ArgumentParser

from pyplanet import __version__ as version
from pyplanet.core import checks
from pyplanet.core.management.color import no_style, color_style


class CommandError(Exception):
	"""
	Exception class indicating a problem while executing a management
	command.
	If this exception is raised during the execution of a management
	command, it will be caught and turned into a nicely-printed error
	message to the appropriate output stream (i.e., stderr); as a
	result, raising this exception (with a sensible description of the
	error) is the preferred way to indicate that something has gone
	wrong in the execution of a command.
	"""
	pass


class SystemCheckError(CommandError):
	"""
	The system check framework detected unrecoverable errors.
	"""
	pass


class CommandParser(ArgumentParser):
	def __init__(self, cmd, **kwargs):
		self.cmd = cmd
		super().__init__(**kwargs)

	def parse_args(self, args=None, namespace=None):
		# Catch missing argument for a better error message
		if (hasattr(self.cmd, 'missing_args_message') and
				not (args or any(not arg.startswith('-') for arg in args))):
			self.error(self.cmd.missing_args_message)
		return super().parse_args(args, namespace)

	def error(self, message):
		if self.cmd._called_from_command_line:
			super().error(message)
		else:
			raise CommandError("Error: %s" % message)


def handle_default_options(options):
	"""
	Include any default options that all commands should accept here
	so that ManagementUtility can handle them before searching for
	user commands.
	"""
	if options.settings:
		os.environ['PYPLANET_SETTINGS_MODULE'] = options.settings
	if options.pythonpath:
		sys.path.insert(0, options.pythonpath)
	if options.pool:
		os.environ['PYPLANET_POOL'] = options.pool


class BaseCommand:
	"""
	The base class from which all management commands ultimately
	derive.

	Use this class if you want access to all of the mechanisms which
	parse the command-line arguments and work out what code to call in
	response; if you don't need to change any of that behavior,
	consider using one of the subclasses defined in this file.

	If you are interested in overriding/customizing various aspects of
	the command-parsing and -execution behavior, the normal flow works
	as follows:

	1. ``pyplanet`` or ``manage.py`` loads the command class
	   and calls its ``run_from_argv()`` method.

	2. The ``run_from_argv()`` method calls ``create_parser()`` to get
	   an ``ArgumentParser`` for the arguments, parses them, performs
	   any environment changes requested by options like
	   ``pythonpath``, and then calls the ``execute()`` method,
	   passing the parsed arguments.

	3. The ``execute()`` method attempts to carry out the command by
	   calling the ``handle()`` method with the parsed arguments; any
	   output produced by ``handle()`` will be printed to standard
	   output and, if the command is intended to produce a block of
	   SQL statements, will be wrapped in ``BEGIN`` and ``COMMIT``.

	4. If ``handle()`` or ``execute()`` raised any exception (e.g.
	   ``CommandError``), ``run_from_argv()`` will  instead print an error
	   message to ``stderr``.

	Thus, the ``handle()`` method is typically the starting point for
	subclasses; many built-in commands and command types either place
	all of their logic in ``handle()``, or perform some additional
	parsing work in ``handle()`` and then delegate from it to more
	specialized methods as needed.

	Several attributes affect behavior at various steps along the way:

	``help``
		A short description of the command, which will be printed in
		help messages.

	``output_transaction``
		A boolean indicating whether the command outputs SQL
		statements; if ``True``, the output will automatically be
		wrapped with ``BEGIN;`` and ``COMMIT;``. Default value is
		``False``. Not used right now, ignored!

	``requires_migrations_checks``
		A boolean; if ``True``, the command prints a warning if the set of
		migrations on disk don't match the migrations in the database.
		Not used currently, ignored!

	``requires_system_checks``
		A boolean; if ``True``, entire PyPlanet project will be checked for errors
		prior to executing the command. Default value is ``True``.
		To validate an individual application's models
		rather than all applications' models, call
		``self.check(app_configs)`` from ``handle()``, where ``app_configs``
		is the list of application's configuration provided by the
		app registry.

	``leave_locale_alone``
		A boolean indicating whether the locale set in settings should be
		preserved during the execution of the command instead of translations
		being deactivated.

		Default value is ``False``.

		Make sure you know what you are doing if you decide to change the value
		of this option in your custom command if it creates database content
		that is locale-sensitive and such content shouldn't contain any
		translations (like it happens e.g. with django.contrib.auth
		permissions) as activating any locale might cause unintended effects.
	"""
	# Metadata about this command.
	help = ''

	# Configuration shortcuts that alter various logic.
	_called_from_command_line = False
	output_transaction = False  # Whether to wrap the output in a "BEGIN; COMMIT;"
	leave_locale_alone = False
	requires_migrations_checks = False
	requires_system_checks = True

	# default arguments required
	arg_pool_required = False
	arg_settings_required = False

	def __init__(self, stdout=None, stderr=None, no_color=False):
		self.stdout = OutputWrapper(stdout or sys.stdout)
		self.stderr = OutputWrapper(stderr or sys.stderr)
		if no_color:
			self.style = no_style()
		else:
			self.style = color_style()
			self.stderr.style_func = self.style.ERROR

	def get_version(self):
		"""
		Return the PyPlanet version, which should be correct for all built-in
		PyPlanet commands. User-supplied commands can override this method to
		return their own version.
		"""
		return version

	def create_parser(self, prog_name, subcommand):
		"""
		Create and return the ``ArgumentParser`` which will be used to
		parse the arguments to this command.
		"""
		parser = CommandParser(
			self, prog='{} {}'.format(os.path.basename(prog_name), subcommand),
			description=self.help or None,
		)
		parser.add_argument('--version', action='version', version=self.get_version())
		parser.add_argument(
			'-v', '--verbosity', action='store', dest='verbosity', default=1,
			type=int, choices=[0, 1, 2, 3],
			help='Verbosity level; 0=minimal output, 1=normal output, 2=verbose output, 3=very verbose output',
		)
		parser.add_argument(
			'--settings',
			help=(
				'The Python path to a settings module, e.g. '
				'"myproject.settings.main". If this isn\'t provided, the '
				'PYPLANET_SETTINGS_MODULE environment variable will be used.'
			),
		)
		parser.add_argument(
			'--pool',
			help=(
				'The PyPlanet pool to use when executing commands. '
				'If this isn\'t provided, the PYPLANET_POOL environment variable will be used.'
				'If the environment variable isn\'t provided, \'default\' will be used.'
			),
			required=self.arg_pool_required,
		)
		parser.add_argument(
			'--pythonpath',
			help='A directory to add to the Python path, e.g. "/home/pyplanetprojects/myproject".',
		)
		parser.add_argument('--traceback', action='store_true', help='Raise on CommandError exceptions')
		parser.add_argument(
			'--no-color', action='store_true', dest='no_color',
			help='Don\'t colorize the command output.',
		)
		self.add_arguments(parser)
		return parser

	def add_arguments(self, parser):
		"""
		Entry point for subclassed commands to add custom arguments.
		:param parser: Parser instance
		:type parser: argparse.ArgumentParser
		"""
		pass

	def print_help(self, prog_name, subcommand):
		"""
		Print the help message for this command, derived from
		``self.usage()``.
		"""
		parser = self.create_parser(prog_name, subcommand)
		parser.print_help()

	def run_from_argv(self, argv):
		"""
		Set up any environment changes requested (e.g., Python path
		and PyPlanet settings), then run this command. If the
		command raises a ``CommandError``, intercept it and print it sensibly
		to stderr. If the ``--traceback`` option is present or the raised
		``Exception`` is not ``CommandError``, raise it.
		"""
		self._called_from_command_line = True
		parser = self.create_parser(argv[0], argv[1])

		options = parser.parse_args(argv[2:])
		cmd_options = vars(options)
		# Move positional args out of options to mimic legacy optparse
		args = cmd_options.pop('args', ())
		handle_default_options(options)
		try:
			self.execute(*args, **cmd_options)
		except Exception as e:
			if options.traceback or not isinstance(e, CommandError):
				raise

			# SystemCheckError takes care of its own formatting.
			if isinstance(e, SystemCheckError):
				self.stderr.write(str(e), lambda x: x)
			else:
				self.stderr.write('{}: {}'.format(e.__class__.__name__, e))
			sys.exit(1)
		finally:
			pass
			# TODO: Close all db connections etc.

	def execute(self, *args, **options):
		"""
		Try to execute this command, performing system checks if needed (as
		controlled by the ``requires_system_checks`` attribute, except if
		force-skipped).
		"""
		if options['no_color']:
			self.style = no_style()
			self.stderr.style_func = None
		if options.get('stdout'):
			self.stdout = OutputWrapper(options['stdout'])
		if options.get('stderr'):
			self.stderr = OutputWrapper(options['stderr'], self.stderr.style_func)

		try:
			if self.requires_system_checks and not options.get('skip_checks'):
				self.check()
			if self.requires_migrations_checks:
				self.check_migrations()
			output = self.handle(*args, **options)
			if output:
				# TODO: Output transactions if self.output_transaction is given!
				self.stdout.write(output)
		finally:
			pass
		return output

	def _run_checks(self, **kwargs):
		return checks.run_checks(**kwargs)

	def check(self, app_configs=None, tags=None, display_num_errors=False,
			  include_deployment_checks=False, fail_level=checks.ERROR):
		"""
		Use the system check framework to validate entire PyPlanet project.
		Raise CommandError for any serious message (error or critical errors).
		If there are only light messages (like warnings), print them to stderr
		and don't raise an exception.
		"""
		all_issues = self._run_checks(
			app_configs=app_configs,
			tags=tags,
			include_deployment_checks=include_deployment_checks,
		)

		header, body, footer = '', '', ''
		visible_issue_count = 0  # excludes silenced warnings

		if all_issues:
			debugs = [e for e in all_issues if e.level < checks.INFO and not e.is_silenced()]
			infos = [e for e in all_issues if checks.INFO <= e.level < checks.WARNING and not e.is_silenced()]
			warnings = [e for e in all_issues if checks.WARNING <= e.level < checks.ERROR and not e.is_silenced()]
			errors = [e for e in all_issues if checks.ERROR <= e.level < checks.CRITICAL and not e.is_silenced()]
			criticals = [e for e in all_issues if checks.CRITICAL <= e.level and not e.is_silenced()]
			sorted_issues = [
				(criticals, 'CRITICALS'),
				(errors, 'ERRORS'),
				(warnings, 'WARNINGS'),
				(infos, 'INFOS'),
				(debugs, 'DEBUGS'),
			]

			for issues, group_name in sorted_issues:
				if issues:
					visible_issue_count += len(issues)
					formatted = (
						self.style.ERROR(str(e))
						if e.is_serious()
						else self.style.WARNING(str(e))
						for e in issues)
					formatted = '\n'.join(sorted(formatted))
					body += '\n{}:\n{}\n'.format(group_name, formatted)

		if visible_issue_count:
			header = 'System check identified some issues:\n'

		if display_num_errors:
			if visible_issue_count:
				footer += '\n'
			footer += 'System check identified {} ({} silenced).'.format(
				'no issues' if visible_issue_count == 0 else
				'1 issue' if visible_issue_count == 1 else
				'%s issues' % visible_issue_count,
				len(all_issues) - visible_issue_count,
			)

		if any(e.is_serious(fail_level) and not e.is_silenced() for e in all_issues):
			msg = self.style.ERROR('SystemCheckError: {}'.format(header)) + body + footer
			raise SystemCheckError(msg)
		else:
			msg = header + body + footer

		if msg:
			if visible_issue_count:
				self.stderr.write(msg, lambda x: x)
			else:
				self.stdout.write(msg)

	def check_migrations(self):
		"""
		Print a warning if the set of migrations on disk don't match the
		migrations in the database.
		"""
		# TODO: Check for migrations. Not yet provided in the cli.
		return

	def handle(self, *args, **options):
		"""
		The actual logic of the command. Subclasses must implement
		this method.
		"""
		raise NotImplementedError('subclasses of BaseCommand must provide a handle() method')


class AppCommand(BaseCommand):
	"""
	A management command which takes one or more installed application labels
	as arguments, and does something with each of them.

	Rather than implementing ``handle()``, subclasses must implement
	``handle_app_config()``, which will be called once for each application.
	"""
	missing_args_message = "Enter at least one application label."

	def add_arguments(self, parser):
		parser.add_argument('args', metavar='app_label', nargs='+', help='One or more application label.')

	def handle(self, *app_labels, **options):
		# TODO: Implement
		return

	def handle_app_config(self, app_config, **options):
		"""
		Perform the command's actions for app_config, an AppConfig instance
		corresponding to an application label given on the command line.
		"""
		raise NotImplementedError(
			"Subclasses of AppCommand must provide"
			"a handle_app_config() method.")


class LabelCommand(BaseCommand):
	"""
	A management command which takes one or more arbitrary arguments
	(labels) on the command line, and does something with each of
	them.

	Rather than implementing ``handle()``, subclasses must implement
	``handle_label()``, which will be called once for each label.

	If the arguments should be names of installed applications, use
	``AppCommand`` instead.
	"""
	label = 'label'
	missing_args_message = 'Enter at least one {}.'.format(label)

	def add_arguments(self, parser):
		parser.add_argument('args', metavar=self.label, nargs='+')

	def handle(self, *labels, **options):
		output = []
		for label in labels:
			label_output = self.handle_label(label, **options)
			if label_output:
				output.append(label_output)
		return '\n'.join(output)

	def handle_label(self, label, **options):
		"""
		Perform the command's actions for ``label``, which will be the
		string as given on the command line.
		"""
		raise NotImplementedError('subclasses of LabelCommand must provide a handle_label() method')


class OutputWrapper(TextIOBase):
	"""
	Wrapper around stdout/stderr
	"""
	@property
	def style_func(self):
		return self._style_func

	@style_func.setter
	def style_func(self, style_func):
		if style_func and self.isatty():
			self._style_func = style_func
		else:
			self._style_func = lambda x: x

	def __init__(self, out, style_func=None, ending='\n'):
		self._out = out
		self.style_func = None
		self.ending = ending

	def __getattr__(self, name):
		return getattr(self._out, name)

	def isatty(self):
		return hasattr(self._out, 'isatty') and self._out.isatty()

	def write(self, msg, style_func=None, ending=None):
		ending = self.ending if ending is None else ending
		if ending and not msg.endswith(ending):
			msg += ending
		style_func = style_func or self.style_func
		self._out.write(style_func(msg))



"""
Converter contrib is managing migrating from another controller.
"""
from pyplanet.contrib.converter.expansion import ExpansionConverter
from pyplanet.contrib.converter.xaseco2 import Xaseco2Converter
from pyplanet.contrib.converter.uaseco import UasecoConverter
from pyplanet.contrib.converter.maniacontrol import ManiacontrolConverter


def get_converter(source_type: str, **options):
	if source_type == 'xaseco2':
		return Xaseco2Converter(**options)
	elif source_type == 'uaseco':
		return UasecoConverter(**options)
	elif source_type == 'maniacontrol':
		return ManiacontrolConverter(**options)
	elif source_type == 'expansion':
		return ExpansionConverter(**options)



from playhouse.shortcuts import model_to_dict

from pyplanet.apps.core.maniaplanet.models import Player

from pyplanet.views.generics.list import ManualListView


class PlayerListView(ManualListView):
	model = Player
	title = 'Players currently on this server'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Buddies'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui

	async def get_data(self):
		return [{'nickname': m.nickname, 'login': m.login, 'is_spectator': 'Yes' if m.flow.is_spectator else 'No'} for m in self.app.instance.player_manager.online]

	async def get_fields(self):
		return [
			{
				'name': 'Nickname',
				'index': 'nickname',
				'sorting': True,
				'searching': True,
				'width': 100,
				'type': 'label'
			},
			{
				'name': 'Login',
				'index': 'login',
				'input': True,
				'sorting': True,
				'searching': True,
				'width': 50,
				'type': 'label'
			},
			{
				'name': 'Is spectator',
				'index': 'is_spectator',
				'sorting': True,
				'searching': False,
				'width': 25,
				'type': 'label'
			},
		]



"""
Dynamic points limit for Shootmania Royal.
"""
import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.apps.core.maniaplanet import callbacks as mp_signals
from pyplanet.contrib.setting import Setting


class DynamicPoints(AppConfig):
	game_dependencies = ['shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.enabled = False

		self.current_limit = None
		self.next_limit = None

		self.setting_enable_dynamic_points = Setting(
			'dynamic_points', 'Enable dynamic points limit', Setting.CAT_FEATURES, type=bool,
			description='Enable the dynamic points limit for Shootmania Royal.',
			default=True, change_target=self.enable_changed
		)
		self.setting_min_points = Setting(
			'min_points', 'Minimum number of points', Setting.CAT_GENERAL, type=int,
			description='The points limit will never go bellow this number of points.',
			default=10, change_target=self.on_changes
		)
		self.setting_max_points = Setting(
			'max_points', 'Maximum number of points', Setting.CAT_GENERAL, type=int,
			description='The points limit will never go above this number of points. Leave at 0 to disable.',
			default=0, change_target=self.on_changes
		)
		self.setting_points_per_player = Setting(
			'per_player_points', 'Points per player', Setting.CAT_GENERAL, type=int,
			description='The points every player should increase the total points limit of the map.',
			default=10, change_target=self.on_changes
		)
		self.setting_change_during_map = Setting(
			'change_during_map', 'Change during map gameplay', Setting.CAT_FEATURES, type=bool,
			description='Enable this to change points during a map gameplay.',
			default=True, change_target=self.on_changes
		)
		self.setting_announce_points_change = Setting(
			'announce_change', 'Announce points change at map start/end', Setting.CAT_FEATURES, type=bool,
			description='The points every player should increase the total points limit of the map.',
			default=True, change_target=self.on_changes
		)

	def enable_changed(self, old, new):
		self.enabled = new

	def is_mode_supported(self, mode):
		return 'royal' in mode.lower()

	async def on_start(self):
		self.context.signals.listen(mp_signals.player.player_connect, self.on_changes)
		self.context.signals.listen(mp_signals.player.player_disconnect, self.on_changes)
		self.context.signals.listen(mp_signals.player.player_info_changed, self.on_changes)
		self.context.signals.listen(mp_signals.map.map_start, self.map_start)

		await self.context.setting.register(
			self.setting_enable_dynamic_points, self.setting_min_points, self.setting_max_points,
			self.setting_points_per_player, self.setting_announce_points_change, self.setting_change_during_map
		)
		self.enabled = await self.setting_enable_dynamic_points.get_value()

		if self.enabled:
			await self.on_changes(enable_announce=True)

	async def on_changes(self, *args, **kwargs):
		if not self.enabled:
			return
		point_per_player = await self.setting_points_per_player.get_value()
		if point_per_player <= 0:
			return

		player_count = self.instance.player_manager.count_players
		total_points = player_count * point_per_player

		if await self.setting_change_during_map.get_value():
			self.current_limit = total_points
			await self.set_limit(total_points, disable_announce='enable_announce' not in kwargs)
			self.next_limit = None
		else:
			self.next_limit = total_points

	async def set_limit(self, new_limit, disable_announce=False):
		min_limit, max_limit, announce = await asyncio.gather(
			self.setting_min_points.get_value(),
			self.setting_max_points.get_value(),
			self.setting_announce_points_change.get_value(),
		)

		if min_limit > 0 and new_limit < min_limit:
			new_limit = min_limit
		if max_limit > 0 and new_limit > max_limit:
			new_limit = max_limit
		if new_limit <= 0 or not new_limit:
			return

		await self.instance.mode_manager.update_settings({
			'S_MapPointsLimit': int(new_limit)
		})
		if announce and not disable_announce:
			await self.instance.chat(
				'$ff0The points limit has been changed to $fff{}$z$s$ff0.'.format(new_limit)
			)

	async def map_start(self, *args, **kwargs):
		if self.next_limit and not await self.setting_change_during_map.get_value():
			self.current_limit = self.next_limit
			await self.set_limit(self.next_limit, disable_announce=kwargs.get('restarted', False))
			self.next_limit = None



#!/usr/bin/env python3
import os
import sys

if __name__ == '__main__':

	# Use Python for the configuration
	os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'python')
	os.environ.setdefault('PYPLANET_SETTINGS_MODULE', 'settings')

	# Use YAML for configuration (comment other options and uncomment the following two lines).
	# os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'yaml')
	# os.environ.setdefault('PYPLANET_SETTINGS_DIRECTORY', 'settings')
	# Will search for two files in the directory: base.yaml and apps.yaml. See http://pypla.net for more information.

	# Use JSON for configuration (comment other options and uncomment the following two lines).
	# os.environ.setdefault('PYPLANET_SETTINGS_METHOD', 'json')
	# os.environ.setdefault('PYPLANET_SETTINGS_DIRECTORY', 'settings')
	# Will search for two files in the directory: base.json and apps.json. See http://pypla.net for more information.

	try:
		from pyplanet.core.management import execute_from_command_line
	except ImportError as exc:
		raise ImportError(
			'Couldn\'t import PyPlanet. Are you sure it\'s installed and '
			'available on your PYTHONPATH environment variable? Did you '
			'forget to activate a virtual environment?'
		) from exc
	execute_from_command_line(sys.argv)



import math
import re
import logging
import pandas as pd
import numpy as np

from asyncio import iscoroutinefunction
from peewee import Field

from pyplanet.utils import style

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.core.db import Model
from pyplanet.views.template import TemplateView

logger = logging.getLogger(__name__)


class ListView(TemplateView):
	"""
	The ListView is an abstract list that uses a database query to show and manipulate the list that is presented to the
	end-user. The ListView is able to automatically manage the searching, ordering and pagination of your query contents.

	The columns could be specified, for each column you can change behaviour, such as searchable and sortable. But also
	custom rendering of the values that will be displayed.

	You can override ``get_fields()``, ``get_actions()``, ``get_query()`` if you need any customization or use a self method
	or variable in one of your properties.

	.. note::

		The design and some behaviour can change in updates of PyPlanet. We aim to provide backward compatibility as much
		as we can. If we are going to break things we will make it deprecated, or if we are in a situation of not having
		enough time to provide a transition time, we are going to create a separate solution (like a second version).

	.. code-block:: python

		class SampleListView(ListView):
			query = Model.select()
			model = Model
			title = 'Select your item'
			fields = [
				{'name': 'Name', 'index': 'name', 'searching': True, 'sorting': True},
				{'name': 'Author', 'index': 'author', 'searching': True, 'sorting': True},
			]
			actions = [
				{
					'name': 'Delete',
					'action': self.action_delete,
					'style': 'Icons64x64_1',
					'substyle': 'Close'
				},
			]

			async def action_delete(self, player, values, instance, **kwargs):
				print('Delete value: {}'.format(instance))

	"""
	query = None
	model = None

	title = None
	icon_style = None
	icon_substyle = None
	fields = []
	actions = []
	buttons = []

	template_name = 'core.views/generics/list.xml'

	single_list = True
	"""Change this to False to have multiple lists open at the same time."""

	def __init__(self, *args, **kwargs):
		self.id = 'pyplanet.views.generics.list.ListView'
		super().__init__(*args, **kwargs)
		self.disable_alt_menu = True
		self.search_text = None
		self.sort_field = None
		self.sort_order = 1
		self.page = 1
		self.count = 0
		self.objects = list()

		self.num_per_page = 20

		self.provide_search = True

		# Setup the receivers.
		self.subscribe('list_button_close', self.close)
		self.subscribe('list_button_refresh', self.refresh)
		self.subscribe('list_button_search', self._search)

		self.subscribe('list_button_first', self._first_page)
		self.subscribe('list_button_prev_10', self._prev_10_pages)
		self.subscribe('list_button_prev', self._prev_page)
		self.subscribe('list_button_next', self._next_page)
		self.subscribe('list_button_next_10', self._next_10_pages)
		self.subscribe('list_button_last', self._last_page)

	@property
	def order(self):
		if self.sort_field and isinstance(self.sort_field, Field):
			if self.sort_order and self.sort_field:
				return self.sort_field
			elif not self.sort_order and self.sort_field:
				return -self.sort_field
		elif self.sort_field:
			return self.sort_field
		return None

	async def handle_catch_all(self, player, action, values, **kwargs):
		# Sorting the column:
		if action.startswith('list_header_'):
			match = re.search('^list_header_([0-9]+)$', action)
			if len(match.groups()) != 1:
				return

			try:
				col = int(match.group(1))
				fields = await self.get_fields()
				field = fields[col]
			except Exception as e:
				logger.error('Got invalid result in list column click: {}'.format(e))
				return

			# Check if sorting is defined + true.
			if 'sorting' not in field or not field['sorting'] or not field['index']:
				return

			# Sort on column
			if isinstance(self.model, Model):
				sort_field = getattr(self.model, field['index'])
				current_field_name = self.sort_field.db_column if self.sort_field else None
				field_name = sort_field.db_column
			else:
				sort_field = field
				current_field_name = self.sort_field['index'] if self.sort_field else None
				field_name = sort_field['index']

			if self.sort_field and current_field_name == field_name:
				if self.sort_order == 1:
					self.sort_order = 0
				else:
					# Unsort. clear sorting
					self.sort_field = None
					self.sort_order = 0
			else:
				self.sort_field = sort_field
				self.sort_order = 1

			# Refresh list
			await self.refresh(player)

		elif action.startswith('list_body_') or action.startswith('list_action_'):
			if action.startswith('list_body_'):
				match = re.search('^list_body_([0-9]+)_([0-9]+)$', action)
				trigger = 'body'
			else:
				match = re.search('^list_action_([0-9]+)_([0-9]+)$', action)
				trigger = 'action'
			if len(match.groups()) != 2:
				return

			try:
				row = int(match.group(1))
				idx = int(match.group(2))
				if trigger == 'body':
					field = (await self.get_fields())[idx]
				else:
					field = (await self.get_actions())[idx]
				action = field['action']
				instance = self.objects[row]
			except Exception as e:
				logger.warning('Got invalid result in list item click: {}'.format(str(e)))
				return

			# Execute action/target method.
			if iscoroutinefunction(action):
				await action(player, values, instance, view=self)
			else:
				action(player, values, instance, view=self)

		elif action.startswith('list_buttons_'):
			match = re.search('^list_buttons_([0-9]+)$', action)

			if len(match.groups()) != 1:
				return

			try:
				button = int(match.group(1))
				field = (await self.get_buttons())[button]
				action = field['action']
			except Exception as e:
				logger.warning('Got invalid result in list item click: {}'.format(str(e)))
				return

			# Execute action/target method.
			if iscoroutinefunction(action):
				await action(player, values, view=self)
			else:
				action(player, values, view=self)

	@property
	def num_pages(self):
		return int(math.ceil(self.count / self.num_per_page))

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])

		if self.single_list:
			# Clear the lock on the player list display.
			player.attributes.set('pyplanet.views.list_displayed', None)

	async def refresh(self, player, *args, **kwargs):
		"""
		Refresh list with current properties for a specific player. Can be used to show new data changes.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		await self.display(player=player)

	async def display(self, player=None):
		"""
		Display list to player.

		:param player: Player login or model instance.
		:type player: str, pyplanet.apps.core.maniaplanet.models.Player
		"""
		login = player.login if isinstance(player, Player) else player
		if not player:
			raise Exception('No player/login given to display the list to!')

		# Check and close other list of user.
		if self.single_list:
			player = player if isinstance(player, Player) \
				else await self.manager.instance.player_manager.get_player(login=login, lock=False)
			other_list = player.attributes.get('pyplanet.views.list_displayed', None)
			if other_list and isinstance(other_list, str):
				# Try to get the other list instance.
				other_manialink = self.manager.instance.ui_manager.get_manialink_by_id(other_list)

				# Close other list and clear lock.
				if isinstance(other_manialink, ListView):
					await other_manialink.close(player)

			# Set lock on player.
			player.attributes.set('pyplanet.views.list_displayed', self.id)

		return await super().display(player_logins=[login])

	async def get_fields(self):
		return self.fields

	async def get_title(self):
		return self.title

	async def get_actions(self):
		return self.actions

	async def get_buttons(self):
		return self.buttons

	async def get_query(self):
		if self.query is not None:
			return self.query
		raise Exception('get_query() or self.query is empty! It should contain query that is not yet executed!')

	async def apply_filter(self, query):
		if not self.search_text:
			return query
		for field in self.fields:
			if 'searching' in field and field['searching']:
				query = query.orwhere(getattr(self.model, field['index']).contains(self.search_text))
		return query

	async def apply_ordering(self, query):
		if not self.order:
			return query
		return query.order_by(self.order)

	async def apply_pagination(self, query):
		# Get count before pagination.
		self.count = await self.model.objects.count(query)
		return query.paginate(self.page, self.num_per_page)

	async def get_object_data(self):
		query = await self.get_query()
		query = await self.apply_filter(query)
		query = await self.apply_ordering(query)
		query = await self.apply_pagination(query)
		self.objects = list(await self.model.execute(query))
		return {
			'objects': self.objects,
			'search': self.search_text,
			'order': self.order,
			'count': self.count,
		}

	async def get_context_data(self):
		context = await super().get_context_data()

		# Add dynamic data from query.
		context.update(await self.get_object_data())

		fields = await self.get_fields()
		actions = await self.get_actions()
		buttons = await self.get_buttons()

		# Process fields + actions (normalize)
		# Calculate positions of fields
		left = 0
		for field in fields:
			field['left'] = left
			left += field['width']
			if 'type' not in field:
				field['type'] = 'label'
			if 'safe' not in field:
				field['safe'] = False

			field['_sort'] = None
			if self.sort_field is not None and field['index'] == self.sort_field['index']:
				field['_sort'] = self.sort_order
		fields_width = int(left)

		left = 0
		for action in actions:
			action['left'] = left
			left += action['width'] if 'width' in action else 5
			if 'type' not in action:
				action['type'] = 'quad'
			if 'safe' not in action:
				action['safe'] = False
		actions_width = int(left)

		right = 215.5
		for button in buttons:
			button['right'] = (right - button['width'] / 2)
			right -= button['width'] + 3

		# Add facts.
		context.update({
			'field_renderer': self._render_field,
			'fields': fields,
			'actions': actions,
			'buttons': buttons,
			'provide_search': self.provide_search,
			'title': await self.get_title(),
			'icon_style': self.icon_style,
			'icon_substyle': self.icon_substyle,
			'search': self.search_text,
			'pages': self.num_pages,
			'page': self.page,
			'fields_width': fields_width,
			'actions_width': actions_width,
		})

		return context

	def _render_field(self, row, field):
		if 'renderer' in field:
			return field['renderer'](row, field)
		if isinstance(row, dict):
			return str(row[field['index']])
		else:
			return str(getattr(row, field['index']))

	async def _search(self, player, _, values, *args, **kwargs):
		search_text = values['list_search_field']
		if len(search_text) > 0 and search_text != 'Search...':
			self.search_text = search_text
		else:
			self.search_text = None
		# Reset page when searching
		self.page = 1
		await self.refresh(player)

	async def _first_page(self, player, *args, **kwargs):
		self.page = 1
		await self.refresh(player)

	async def _last_page(self, player, *args, **kwargs):
		self.page = self.num_pages
		await self.refresh(player)

	async def _next_page(self, player, *args, **kwargs):
		if self.page + 1 <= self.num_pages:
			self.page += 1
			await self.refresh(player)

	async def _next_10_pages(self, player, *args, **kwargs):
		if self.page + 10 <= self.num_pages:
			self.page += 10
		else:
			self.page = self.num_pages
		await self.refresh(player)

	async def _prev_page(self, player, *args, **kwargs):
		if self.page - 1 > 0:
			self.page -= 1
			await self.refresh(player)

	async def _prev_10_pages(self, player, *args, **kwargs):
		if self.page - 10 > 0:
			self.page -= 10
		else:
			self.page = 1
		await self.refresh(player)


class ManualListView(ListView):
	"""
	The ManualListView will act as a ListView, but not based on a model or query.
	"""

	def __init__(self, data=None, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.objects_raw = data

	async def get_data(self):
		"""
		Override this method, return a list with dictionaries inside.
		"""
		if not self.objects_raw:
			raise NotImplementedError
		return self.objects_raw

	async def get_object_data(self):
		frame = pd.DataFrame(await self.get_data())
		frame = await self.apply_filter(frame)
		frame = await self.apply_ordering(frame)
		self.count = len(frame)
		frame = await self.apply_pagination(frame)
		self.objects = frame.to_dict('records')
		return {
			'objects': self.objects,
			'search': self.search_text,
			'order': self.order,
			'count': self.count,
		}

	async def apply_filter(self, frame):
		if not self.search_text:
			return frame
		query = list()
		for field in await self.get_fields():
			if 'searching' in field and field['searching']:
				if 'search_strip_styles' in field and field['search_strip_styles']:
					query.append(
						frame[field['index']].apply(lambda x: self.search_text.lower() in style.style_strip(str(x).lower()) if x else False)
					)
				else:
					query.append(
						frame[field['index']].apply(lambda x: self.search_text.lower() in str(x).lower() if x else False)
					)
		if query:
			query = np.logical_or.reduce(query)
			return frame.loc[query]
		return frame

	async def apply_ordering(self, frame):
		if self.sort_field:
			return frame.sort_values(self.sort_field['index'], ascending=bool(self.sort_order))
		return frame

	async def apply_pagination(self, frame):
		return frame[(self.page - 1) * self.num_per_page:self.page * self.num_per_page]



"""
Custom Call Views.
"""
import asyncio
import logging

from xmlrpc.client import Fault

from pyplanet.views import TemplateView
from pyplanet.views.generics import ManualListView


class CallMenuView(ManualListView):
	title = 'Custom XMLRPC Call List'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Profile'

	def __init__(self, app, player, type='native'):
		"""
		:param app: App config instance.
		:param player: Player instance.
		:param type: Type of methods, can be 'native' or 'scripted'.
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type type: str
		"""
		super().__init__()
		self.manager = app.context.ui
		self.app = app
		self.player = player

		self.child = None

		self.type = type
		self.cache = list()

	async def get_buttons(self):
		return list()
		# return [dict(
		# 	title='Custom method',
		# 	width=35,
		# 	action=self.custom_call,
		# )]

	async def custom_call(self, player, values, **kwargs):
		pass

	async def get_data(self):
		if self.cache:
			return self.cache

		methods = await self.app.instance.gbx('system.listMethods')
		method_signatures, method_helps = await asyncio.gather(
			self.app.instance.gbx.multicall(
				*[self.app.instance.gbx('system.methodSignature', m) for m in methods]
			),
			self.app.instance.gbx.multicall(
				*[self.app.instance.gbx('system.methodHelp', m) for m in methods]
			)
		)

		self.cache = [
			dict(
				key=method,
				name=method,
				inputs=','.join(method_signature[0][1:]) if len(method_signature) and len(method_signature[0]) > 1 else '-',
				# inputs=method_signature,
				inputs_raw=method_signature[0][1:] if len(method_signature) and len(method_signature[0]) > 1 else '-',
				signature=method_signature,
				output=method_signature[0][0] if len(method_signature) and len(method_signature[0]) else '-',
				help=method_help,
				type=self.type
			)
			for method, method_signature, method_help in zip(methods, method_signatures, method_helps)
		]

		return self.cache

	async def open_call_window(self, player, values, row, **kwargs):
		if self.child:
			return

		# Getting method information.
		method = dict(
			name=row['key'],
			method=row['key'],
			signature=row['signature'],
			inputs=row['inputs_raw'],
			output=row['output'],
			help=row['help'],
			type=row['type'],
		)

		# Show edit view.
		self.child = CallActionView(self, self.player, method)
		await self.child.display()
		await self.child.wait_for_response()
		await self.child.destroy()
		self.child = None

	async def display(self, **kwargs):
		kwargs['player'] = self.player
		return await super().display(**kwargs)

	async def get_fields(self):
		return [
			{
				'name': 'Name',
				'index': 'name',
				'sorting': True,
				'searching': True,
				'width': 40,
				'type': 'label'
			},
			{
				'name': 'Input(s)',
				'index': 'inputs',
				'sorting': True,
				'searching': False,
				'width': 35,
				'type': 'label'
			},
			{
				'name': 'Output',
				'index': 'output',
				'sorting': True,
				'searching': False,
				'width': 35,
				'type': 'label'
			},
			{
				'name': 'Help',
				'index': 'help',
				'sorting': False,
				'searching': True,
				'width': 75,
				'type': 'label'
			},
			{
				'name': 'Type',
				'index': 'type',
				'sorting': True,
				'searching': False,
				'width': 15,
				'type': 'label'
			},
		]

	async def get_actions(self):
		return [
			{
				'name': 'Call',
				'type': 'label',
				'text': '$fff &#xf040; Call',
				'width': 11,
				'action': self.open_call_window,
				'safe': True
			},
		]


class CallActionView(TemplateView):
	"""
	Call Action View.
	"""
	template_name = 'core.pyplanet/call/call.xml'

	def __init__(self, parent, player, method):
		"""
		Initiate child call view.

		:param parent: Parent view.
		:param player: Player instance.
		:param method: Method dictionary.
		:type parent: pyplanet.view.base.View
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:type method: dict
		"""
		super().__init__(parent.manager)
		self.parent = parent
		self.player = player
		self.method = method

		self.response_future = asyncio.Future()

		self.subscribe('button_close', self.close)
		self.subscribe('button_execute', self.execute)
		self.subscribe('button_cancel', self.close)

	async def display(self, **kwargs):
		await super().display(player_logins=[self.player.login])

	async def get_context_data(self):
		context = await super().get_context_data()

		context['method'] = self.method

		context['title'] = 'Call \'{}\''.format(self.method['name'])
		context['icon_style'] = 'Icons128x128_1'
		context['icon_substyle'] = 'Profile'

		return context

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])
		self.response_future.set_result(None)
		self.response_future.done()

	async def wait_for_response(self):
		return await self.response_future

	def convert_type(self, value, requested_type):
		if requested_type == 'boolean':
			return bool(value)
		elif requested_type == 'array':
			return value.splitlines()
		elif requested_type == 'double':
			return float(value)
		elif requested_type == 'int':
			return int(value)
		return value

	def format_output(self, raw_output):
		if self.method['output'] == 'array':
			return '\n'.join(raw_output)
		elif self.method['output'] == 'struct':
			return '\n'.join(['\'{}\': \'{}\''.format(key, value) for key, value in raw_output.items()])
		return raw_output

	async def execute(self, player, action, values, *_, **__):
		self.method['last_input'] = values['call_value_field']
		input_values = str(values['call_value_field']).splitlines()
		args = list()

		if len(self.method['inputs']) > 0:
			for idx, input in enumerate(self.method['inputs']):
				if len(input_values) <= idx:
					await self.parent.app.instance.chat('$ff0Warning: Not enough inputs given for the call!')
					continue

				args.append(self.convert_type(input_values[idx], input))

		# Execute call.
		try:
			logging.getLogger(__name__).warning('Executing GBX in-game: {}, {}'.format(self.method['method'], args))
			result = await self.parent.app.instance.gbx(self.method['method'], *args)
			self.method['result'] = self.format_output(result)

			# Refresh with results.
			await self.display()
		except Fault as e:
			logging.getLogger(__name__).warning('Executing GBX in-game, fault thrown: {}'.format(str(e)))
			self.method['error'] = '{}: {}'.format(e.faultCode, e.faultString)

			# Refresh with error.
			await self.display()
		except Exception as e:
			logging.getLogger(__name__).warning('Executing GBX in-game, error thrown: {}'.format(str(e)))
			await self.hide([player.login])
			self.response_future.set_result(False)
			self.response_future.done()



"""
Player Admin methods and functions.
"""
from pyplanet.contrib.command import Command


class PyPlanetAdmin:
	def __init__(self, app):
		"""
		:param app: App instance.
		:type app: pyplanet.apps.contrib.admin.app.Admin
		"""
		self.app = app
		self.instance = app.instance

	async def on_start(self):
		await self.instance.permission_manager.register('reboot', 'Reboot PyPlanet pool instance', app=self.app, min_level=3)

		await self.instance.command_manager.register(
			Command(command='reboot', target=self.reboot_pool, perms='admin:reboot', admin=True, description='Reboots PyPlanet.'),
		)

	async def reboot_pool(self, player, data, **kwargs):
		exit(50)



import asynctest

from pyplanet.core import Controller
from pyplanet.core.events import Signal, Callback


class TestCallbacks(asynctest.TestCase):
	async def test_registering(self):
		instance = Controller.prepare(name='default').instance

		test1 = Callback(
			call='SampleCall',
			code='sample_call',
			namespace='tests',
			target=self.handle_sample
		)

		test1_comp = instance.signals.get_callback('SampleCall')

		assert test1.raw_signal == test1_comp

	async def test_listening(self):
		test1 = Callback(
			call='SampleCall',
			code='sample_call',
			namespace='tests',
			target=self.handle_sample
		)

		self.got_sync = 0
		self.got_async = 0
		self.got_handle = 0
		self.got_glue = 0

		test1.register(self.sync_listener)
		test1.register(self.async_listener)

		await test1.raw_signal.send(dict(glue=False), raw=False)
		await test1.raw_signal.send_robust(dict(glue=False), raw=False)

		assert self.got_async == 2
		assert self.got_sync == 2
		assert self.got_glue == 0
		assert self.got_handle == 2

	####################################################################################################################

	def sync_listener(self, *args, **kwargs):
		self.got_sync += 1

	async def async_listener(self, *args, **kwargs):
		self.got_async += 1

	async def glue(self, *args, **kwargs):
		self.got_glue += 1
		return dict(glue=True)

	async def handle_sample(self, *args, **kwargs):
		self.got_handle += 1
		return dict(handle=True)



import logging
import asyncio

from pyplanet.apps.config import AppConfig
from pyplanet.contrib.command import Command
from pyplanet.utils import style

from pyplanet.apps.contrib.queue.list import QueueList
from pyplanet.apps.contrib.queue.view import QueueView

from pyplanet.apps.core.maniaplanet import callbacks as mp_signals

logger = logging.getLogger(__name__)


class Queue(AppConfig):
	game_dependencies = ['trackmania', 'shootmania']
	app_dependencies = ['core.maniaplanet']

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self.list = QueueList()
		self.list.change_hook = self.change_hook
		self.widget = QueueView(self)
		self.has_password = False

		self.context.signals.listen(mp_signals.player.player_info_changed, self.player_change)
		self.context.signals.listen(mp_signals.player.player_connect, self.player_connect)
		self.context.signals.listen(mp_signals.player.player_disconnect, self.player_disconnect)
		self.context.signals.listen(mp_signals.other.server_password, self.password_change)

		self.context.signals.listen(mp_signals.player.player_enter_player_slot, self.slot_change)
		self.context.signals.listen(mp_signals.player.player_enter_spectator_slot, self.slot_change)

	async def on_start(self):
		# Register permission & command for managing queue.
		await self.instance.permission_manager.register('manage_queue', 'Clear or shuffle the queue', app=self, min_level=2)
		await self.instance.command_manager.register(
			Command('queue', target=self.command_queue_list),
			Command('clear', namespace='queue', target=self.command_queue_clear, admin=True, perms='queue:manage_queue'),
			Command('shuffle', namespace='queue', target=self.command_queue_shuffle, admin=True, perms='queue:manage_queue'),
		)

		self.has_password = self.instance.game.server_password is not None and len(self.instance.game.server_password) > 0

		await self.init_queue()

	async def on_stop(self):
		"""
		On stop will make sure the spectators can change to player slots again.
		"""
		try:
			await self.instance.gbx.multicall(*[
				self.instance.gbx('ForceSpectator', l, 0) for l in self.instance.player_manager.online_logins
			])
		except:
			pass

	async def init_queue(self):
		"""
		Init/enable the queue

		:return:
		"""
		# Check for server password, if set, ignore and disable app.
		if self.has_password:
			logger.debug('Not loading queue! Player password has been set!')
			return

		# Make sure the spectators can't switch to player, and players can't join directly as player.
		# Also make sure the widget is displayed to the spectators.
		calls = []
		logins = []
		for player in self.instance.player_manager.online:
			if player.flow.is_spectator:
				calls.append(self.instance.gbx('ForceSpectator', player.login, 1))
				logins.append(player.login)
		if len(calls) > 0:
			await asyncio.gather(
				self.instance.gbx.multicall(*calls),
				self.widget.display(player_logins=logins)
			)

	async def disable_queue(self):
		"""
		Disable the queue widget and functions.

		:return:
		"""
		logger.debug('Stopping the queue app! User password set?!')

		await self.list.clear()

		# Change the force spec state on players (unforce state).
		try:
			await self.instance.gbx.multicall(*[
				self.instance.gbx('ForceSpectator', l, 0) for l in self.instance.player_manager.online_logins
			])
		except:
			pass

		# Hide all widgets.
		await self.widget.hide()

	async def password_change(self, password, kind, **kwargs):
		if kind == 'player' and password is not None:
			self.has_password = True
			await self.disable_queue()

		if self.has_password and kind == 'player' and password is None:
			self.has_password = False
			await self.init_queue()

	async def player_change(self, player, **_):
		"""
		Player Change Signal.

		:param player: player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:return:
		"""
		if self.has_password:
			return

		if player.flow.is_spectator and player.flow.has_player_slot:
			# Release the player slot to the game.
			try:
				await self.instance.gbx('SpectatorReleasePlayerSlot', player.login)
			except:
				# Most likely the player just left the server. Ignore and return the method.
				return
			logger.debug('Release player slot of player {}'.format(player.login))

		# Display widget if spectator.
		if player.flow.is_spectator:
			await self.widget.display(player_logins=[player.login])
		else:
			await self.widget.hide(player_logins=[player.login])

	async def player_connect(self, player, is_spectator, **_):
		"""
		Player connect signal

		:param player: Player instance
		:param is_spectator: Is player spectator?
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		if self.has_password:
			return

		if player.flow.is_server or player.flow.is_referee:
			return

		# Force player into spectator when joining as player (only when people are in the queue).
		if not is_spectator and await self.list.count() > 0:
			await self.instance.gbx.multicall(
				self.instance.gbx('ForceSpectator', player.login, 1),
				self.instance.chat(
					'$39fYou are currently forced into spectator. Use the Queue widget to obtain a player spot.', player
				)
			)
		await self.slot_change()

	async def player_disconnect(self, player, **_):
		"""
		Player Disconnect, remove from queue if in and reindex queue.

		:param player:
		:return:
		"""
		await self.exit_queue(player)
		await self.slot_change()

	async def slot_change(self, *args, **kwargs):
		"""
		Any slot changes will be captured here. We will investigate the free player spots and put a player into player
		mode if any player is in the queue and free spaces are available.
		"""
		if self.has_password:
			return

		max_players = self.instance.game.server_max_players
		num_players = self.instance.player_manager.count_players
		free_spots = max_players - num_players

		# Stop if no free spots.
		if free_spots <= 0:
			return

		with await self.list.lock:
			num_queue = await self.list.count()

			# Stop if no players are in the queue.
			if num_queue == 0:
				return

			# Pop the next player from the queue.
			next_player = await self.list.pop()
			if not next_player:
				return

			# Force player into player slot.
			await self.instance.gbx('ForceSpectator', next_player.login, 2)
			await self.instance.gbx('ForceSpectator', next_player.login, 0)

			await self.instance.chat(
				'$39fPlayer $fff{} $z$s$49f has been released from the queue! Have fun!.'.format(next_player.nickname)
			)

		# Make sure the widgets are updated for all other spectators.
		logins = [p.login for p in self.instance.player_manager.online if p.flow.is_spectator]
		await self.widget.display(player_logins=logins)

	async def change_hook(self, action=None, entity=None, *args, **kwargs):
		# Check the action of the event.
		if action in ['push', 'pop', 'remove', 'clear', 'shuffle']:
			# Reload the widget to all the players that should see it (all spectators).
			logins = [p.login for p in self.instance.player_manager.online if p.flow.is_spectator]
			await self.widget.display(player_logins=logins)

	async def enter_queue(self, player):
		"""
		Let the given player enter the queue.

		:param player: player instance
		:return: the given position or None if failed or already in the queue.
		"""
		return await self.list.push(player)

	async def exit_queue(self, player):
		"""
		Remove the player from the queue.

		:param player: player instance
		:return: Boolean determinating if the player has been removed.
		"""
		return await self.list.remove(player)

	async def command_queue_list(self, player, *args, **kwargs):
		"""
		/queue command

		:param player: Player instance
		"""
		nicknames = [style.style_strip(p.nickname, style.STRIP_SIZES, style.STRIP_COLORS) for p in await self.list.copy()]
		if nicknames:
			await self.instance.chat(
				'$39fCurrent queue: $fff{}'.format('$39f,$fff '.join(nicknames)), player
			)
		else:
			await self.instance.chat(
				'$39fThere is nobody in the waiting queue!', player
			)

	async def command_queue_clear(self, player, *args, **kwargs):
		"""
		//queue clear

		:param player: Player instance
		"""
		await self.list.clear()
		await asyncio.gather(
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has cleared the waiting queue!'.format(player.nickname)),
			self.slot_change()
		)

	async def command_queue_shuffle(self, player, *args, **kwargs):
		"""
		//queue shuffle

		:param player: Player instance
		"""
		await self.list.shuffle()
		await asyncio.gather(
			self.instance.chat('$ff0Admin $fff{}$z$s$ff0 has shuffled the waiting queue!'.format(player.nickname)),
			self.slot_change()
		)



import io
import struct
from asyncio import iscoroutinefunction

import aiofiles


class GbxException(BaseException):
	"""
	Exception with parsing the Gbx file.
	"""
	pass


class _LookBackUtils:
	PREDEFINED_STRINGS = {
		11: 'Valley',
		12: 'Canyon',
		13: 'Lagoon',
		17: 'TMCommon',
		26: 'Stadium',
		202: 'Storm',
		299: 'SMCommon',
		10003: 'Common',
	}

	def __init__(self, buffer):
		self.buffer = buffer
		self.store = list()
		self.version = None

	async def read_string(self):
		length, = struct.unpack('<L', await self.buffer.read(4))
		return struct.unpack('<{}s'.format(length), await self.buffer.read(length))[0].decode()

	async def read_lookback_string(self):
		if self.version is None:
			# We should see the lookback version right now.
			self.version, = struct.unpack('<L', await self.buffer.read(4))
		# Get the index.
		idx, = struct.unpack('<L', await self.buffer.read(4))
		if idx == 0:
			return None

		# Check if this will be the first occurrence.
		if idx & 0xc0000000 != 0 and idx & 0x3fffffff == 0:
			value = await self.read_string()
			self.store.append(value)
			return value

		# Check if idx is telling us that it's empty.
		if idx == 0xffffffff:
			return ''

		# Check if it's a predefined value.
		if idx & 0x3fffffff == idx:
			return self.PREDEFINED_STRINGS[idx]

		# Get from local store.
		idx &= 0x3fffffff
		if idx - 1 >= len(self.store):
			raise GbxException('String not found in lookback list!. Offset: {}'.format(await self.buffer.tell))
		return self.store[idx - 1]

	def reset(self):
		if self.version:
			self.store = list()
			self.version = None


class _AsyncBufferProxy:
	def __init__(self, buffer):
		"""
		:param buffer: Buffer
		:type buffer: io.BufferedIOBase
		"""
		self.buffer = buffer

	async def read(self, size=1):
		return self.buffer.read(size)

	async def seek(self, offset, whence=io.SEEK_CUR):
		return self.buffer.seek(offset, whence)

	async def tell(self):
		return self.buffer.tell()


class GbxParser:
	"""
	Async GBX Map Information Parser.

	Author: Toffe.
	"""

	def __init__(self, file=None, buffer=None):
		"""
		Initiate a parser with either a file path or buffer.

		:param file: File path.
		:param buffer: Buffer
		:type file: str
		"""
		super().__init__()
		if file and not isinstance(file, str):
			raise Exception('File should be a string, pointing to the file you want to load.')
		if not file and not buffer:
			raise Exception('File or buffer is required!')
		self.file = file

		if buffer:
			if iscoroutinefunction(buffer.read):
				self.buffer = buffer
			else:
				self.buffer = _AsyncBufferProxy(buffer)
		else:
			self.buffer = _AsyncBufferProxy(buffer)
		self.strings = _LookBackUtils(self.buffer)

		self.result = dict()

		self.header = None
		self.header_xml = None
		self.header_length = 0
		self.header_chunk_count = 0
		self.header_chunks = dict()

		self.parse_thumb = False
		self.parse_header_xml = False

	async def seek(self, offset):
		"""
		We need to override the second param to move from the current position.

		:param offset: offset to move away.
		:type offset: int
		"""
		return await self.buffer.seek(offset, io.SEEK_CUR)

	async def parse(self, thumb=False, header_xml=False):
		self.parse_thumb = thumb
		self.parse_header_xml = header_xml

		if self.file:
			async with aiofiles.open(self.file, mode='rb') as self.buffer:
				self.strings = _LookBackUtils(self.buffer)
				return await self.__parse()
		elif self.buffer:
			return await self.__parse()
		raise Exception('No buffer or file given at init.')

	async def __parse(self):
		# Skip until class reference.
		await self.seek(9)
		# Read class ID.
		class_id, = struct.unpack('<I', await self.buffer.read(4))
		if class_id != ((0x3 << 24) | (0x43 << 12)):
			raise GbxException('Gbx file has no valid parser, only maps are supported right now.')

		self.result.update(await self.__parse_header())

		return self.result

	async def __parse_header(self):
		self.header_length, = struct.unpack('<I', await self.buffer.read(4))
		self.header_chunk_count, = struct.unpack('<I', await self.buffer.read(4))

		self.header_chunks = dict()
		self.header = dict()

		# Save header data from binary.
		for nr in range(self.header_chunk_count):
			chunk_id, = struct.unpack('<I', await self.buffer.read(4))
			chunk_size, = struct.unpack('<I', await self.buffer.read(4))
			self.header_chunks[chunk_id] = chunk_size & ~0x80000000

		# Parse all header chunks.
		for chunk_id, chunk_size in self.header_chunks.items():
			self.strings.reset()
			self.header.update(await self.__parse_chunk(chunk_id, chunk_size))

		return self.header

	async def __parse_chunk(self, chunk_id, chunk_size):
		if chunk_id == 0x03043002:
			version, = struct.unpack('<B', await self.buffer.read(1))
			await self.seek(4)
			time_bronze, time_silver, time_gold, time_author = struct.unpack('<LLLL', await self.buffer.read(16))
			price, is_multilap, map_type = struct.unpack('<LLL', await self.buffer.read(12))
			is_multilap = True if is_multilap == 1 else False
			await self.seek(4)
			author_score, editor = struct.unpack('<LL', await self.buffer.read(8))
			editor = 'simple' if editor == 1 else 'advanced'
			await self.seek(4)
			checkpoints, laps = struct.unpack('<LL', await self.buffer.read(8))
			return dict(
				time_bronze=time_bronze, time_silver=time_silver, time_gold=time_gold, time_author=time_author,
				price=price, is_multilap=is_multilap, map_type=map_type, author_score=author_score, editor=editor,
				checkpoints=checkpoints, laps=laps
			)

		elif chunk_id == 0x03043003:
			version, = struct.unpack('<B', await self.buffer.read(1))
			uid = await self.strings.read_lookback_string()
			environment = await self.strings.read_lookback_string()
			author_login = await self.strings.read_lookback_string()
			name = await self.strings.read_string()
			await self.seek(5)
			await self.strings.read_string() # Unknown, mostly empty.
			mood = await self.strings.read_lookback_string()
			decoration_env_id = await self.strings.read_lookback_string()
			decoration_env_author = await self.strings.read_lookback_string()
			await self.seek(4*4+16)
			map_type = await self.strings.read_string()
			map_style = await self.strings.read_string()
			await self.seek(9)
			title_id = await self.strings.read_lookback_string()
			return dict(
				uid=uid, environment=environment, author_login=author_login, name=name, mood=mood,
				decoration_env_id=decoration_env_id, decoration_env_author=decoration_env_author,
				map_type=map_type, map_style=map_style, title_id=title_id
			)

		elif chunk_id == 0x03043004:
			version, = struct.unpack('<B', await self.buffer.read(1))
			await self.seek(chunk_size - 1)

		elif chunk_id == 0x03043005:
			self.header_xml = await self.strings.read_string()

		elif chunk_id == 0x03043007:
			has_thumb = bool(struct.unpack('<L', await self.buffer.read(4))[0])
			comment = None
			thumb = None
			if has_thumb:
				thumb_size, = struct.unpack('<L', await self.buffer.read(4))
				await self.seek(15) # Skip XML thumb tag.
				if self.parse_thumb:
					thumb = struct.unpack('<{}s'.format(thumb_size), await self.buffer.read(thumb_size))[0].decode()
				else:
					await self.seek(thumb_size)
				await self.seek(16 + 10) # </Thumbnail.jpg></Comments>

				comment_size, = struct.unpack('<L', await self.buffer.read(4))
				if comment_size > 0:
					comment = struct.unpack('<{}s'.format(comment_size), await self.buffer.read(comment_size))[0].decode()
				await self.seek(11) # </Comments>
			else:
				await self.seek(chunk_size - 4)

			return dict(has_thumb=has_thumb, thumb=thumb, comment=comment)

		elif chunk_id == 0x03043008:
			version, = struct.unpack('<L', await self.buffer.read(4))
			author_version, = struct.unpack('<L', await self.buffer.read(4))
			author_login = await self.strings.read_string()
			author_nickname = await self.strings.read_string()
			author_zone = await self.strings.read_string()
			author_extra = await self.strings.read_string()
			return dict(
				author_version=author_version, author_login=author_login, author_nickname=author_nickname,
				author_zone=author_zone, author_extra=author_extra
			)
		return dict()



from .database import Database, Proxy
from .registry import Registry
from .migrator import Migrator
from .model import Model, TimedModel

__all__ = [
	'Database',
	'Registry',
	'Proxy',
	'Migrator',
	'Model',
	'TimedModel',
]



"""
The database class in this file holds the engine and state of the database connection. Each instance has one specific
database instance running.
"""
import contextlib
import importlib
import logging
import peewee
import peewee_async

from pyplanet.core.exceptions import ImproperlyConfigured
from .registry import Registry
from .migrator import Migrator

Proxy = peewee.Proxy()


class Database:
	def __init__(self, engine_cls, instance, *args, **kwargs):
		"""
		Initiate database.

		:param engine_cls: Engine class
		:param instance: Instance of the app.
		:param args: *
		:param kwargs: **
		:type instance: pyplanet.core.instance.Instance
		"""
		self.engine = engine_cls(*args, **kwargs)
		self.instance = instance
		self.migrator = Migrator(self.instance, self)
		self.registry = Registry(self.instance, self)
		self.objects = peewee_async.Manager(self.engine, loop=self.instance.loop)

		# Don't allow any sync code.
		if hasattr(self.engine, 'allow_sync'):
			self.engine.allow_sync = False

		Proxy.initialize(self.engine)

	@classmethod
	def create_from_settings(cls, instance, conf):
		try:
			engine_path, _, cls_name = conf['ENGINE'].rpartition('.')
			db_name = conf['NAME']
			db_options = conf['OPTIONS'] if 'OPTIONS' in conf and conf['OPTIONS'] else dict()

			# FIX for #331. Replace utf8 by utf8mb4 in the mysql driver encoding.
			if conf['ENGINE'] == 'peewee_async.MySQLDatabase' and 'charset' in db_options and db_options['charset'] == 'utf8':
				logging.info('Forcing to use \'utf8mb4\' instead of \'utf8\' for the MySQL charset option! (Fix #331).')
				db_options['charset'] = 'utf8mb4'

			# We will try to load it so we have the validation inside this class.
			engine = getattr(importlib.import_module(engine_path), cls_name)
		except ImportError:
			raise ImproperlyConfigured('Database engine doesn\'t exist!')
		except Exception as e:
			raise ImproperlyConfigured('Database configuration isn\'t complete or engine could\'t be found!')

		return cls(engine, instance, db_name, **db_options)

	@contextlib.contextmanager
	def __fake_allow_sync(self):
		try:
			yield
		except:
			raise

	def allow_sync(self, *args, **kwargs):
		"""
		Wrapper around engine allow_sync to allow failover when no async driver.

		:param args:
		:param kwargs:
		:return:
		"""
		if hasattr(self.engine, 'allow_sync'):
			return self.engine.allow_sync()
		return self.__fake_allow_sync()

	async def connect(self):
		self.engine.connect()
		logging.info('Database connection established!')

	async def initiate(self):
		# Create the migration table.
		from .models import migration
		with self.allow_sync():
			migration.Migration.create_table(True)

		# Execute checks.
		with self.allow_sync():
			await self.migrator.check()

		# Migrate all models.
		with self.allow_sync():
			await self.migrator.migrate()

	async def drop_tables(self):
		from .models import migration
		with self.allow_sync():
			try:
				migration.Migration.drop_table(True, True)
			except:
				pass

			restart = 0
			while restart < 5:
				for name, (_, _, model) in self.registry.models.items():
					try:
						model.drop_table(True, True)
					except:
						restart += 1



"""
The module and class MXKarma is the communication between the API and the Karma class.
"""
import datetime

from pyplanet.contrib.setting import Setting

from pyplanet.apps.contrib.karma.mxkarmaapi import MXKarmaApi


class MXKarma:
	"""
	The MX Karma sub-app of the Karma app.
	"""

	def __init__(self, app):
		self.app = app
		self.api = MXKarmaApi(self)

		self.current_count = 0
		self.current_average = 0.0
		self.current_start = datetime.datetime.now()
		self.current_votes = None

		self.setting_mx_karma = Setting(
			'mx_karma', 'Enable MX Karma', Setting.CAT_BEHAVIOUR, type=bool,
			description='Enabling MX Karma will provide you with global karma information from ManiaExchange.',
			default=False, change_target=self.reload_settings
		)

		self.setting_mx_karma_key = Setting(
			'mx_karma_key', 'MX Karma API Key', Setting.CAT_BEHAVIOUR, type=str,
			description='Enabling MX Karma will provide you with global karma information from ManiaExchange.',
			default=None, change_target=self.reload_settings
		)

	async def reload_settings(self, *args, **kwargs):
		"""
		Reload the settings from the PyPlanet settings storage and reload/restart the session.

		:param args: args
		:param kwargs: kwargs
		"""
		enabled = await self.setting_mx_karma.get_value()
		key = await self.setting_mx_karma_key.get_value()

		if enabled is True and key is not None:
			await self.api.create_session()
			await self.api.start_session()
		else:
			await self.api.close_session()

	async def determine_vote(self, vote):
		"""
		Convert a local vote to a MX vote value.

		:param vote: vote value of PyPlanet karma app.
		:return: MX Karma vote value
		"""
		mx_vote = 0
		if vote == -0.5:
			mx_vote = 25
		elif vote == 0:
			mx_vote = 50
		elif vote == 0.5:
			mx_vote = 75
		elif vote == 1:
			mx_vote = 100

		return mx_vote

	async def handle_rating(self, rating, importvotes=True):
		"""
		Handle the event of a rating.

		:param rating: Rating dict
		:param importvotes: Import boolean
		:return:
		"""
		if rating is not None:
			self.current_count = rating['votecount']
			self.current_average = rating['voteaverage']
			self.current_votes = rating['votes']

			if not importvotes:
				return

			import_votes = []
			for vote in self.app.current_votes:
				login = vote.player.login
				score = vote.score
				if vote.expanded_score is not None:
					score = vote.expanded_score

				if not any(mx['login'] == login for mx in self.current_votes):
					import_votes.append({'login': login, 'nickname': vote.player.nickname, 'vote': await self.determine_vote(score)})

			if len(import_votes) > 0:
				if await self.api.save_votes(map=self.app.instance.map_manager.current_map, is_import=True, votes=import_votes):
					rating = await self.api.get_map_rating(self.app.instance.map_manager.current_map)
					await self.handle_rating(rating, importvotes=False)

					if len(self.current_votes) != len(import_votes):
						self.current_votes = import_votes
		else:
			self.current_count = 0
			self.current_average = 0.0
			self.current_votes = None

	async def on_start(self):
		"""
		On start of module.
		"""
		await self.app.context.setting.register(
			self.setting_mx_karma, self.setting_mx_karma_key
		)

		if await self.setting_mx_karma.get_value() is False or await self.setting_mx_karma_key.get_value() is None:
			return

		self.current_start = datetime.datetime.now()

		await self.api.create_session()
		await self.api.start_session()

		rating = await self.api.get_map_rating(self.app.instance.map_manager.current_map)
		await self.handle_rating(rating)

	async def on_stop(self):
		"""
		On stop of module.
		"""
		if await self.setting_mx_karma.get_value() is False or await self.setting_mx_karma_key.get_value() is None:
			return

		await self.api.close_session()

	async def player_connect(self, player):
		"""
		Get player rating on connection.

		:param player: player instance
		"""
		rating = await self.api.get_map_rating(self.app.instance.map_manager.current_map, player.login)

	async def map_begin(self, map):
		"""
		On map begin, load karma.

		:param map: map instance
		"""
		if await self.setting_mx_karma.get_value() is False or await self.setting_mx_karma_key.get_value() is None:
			return

		if not self.api.activated:
			return

		self.current_start = datetime.datetime.now()

		rating = await self.api.get_map_rating(map)
		await self.handle_rating(rating)

	async def map_end(self, map):
		"""
		Map end, save votes to API if needed.

		:param map: map instance.
		"""
		if await self.setting_mx_karma.get_value() is False or await self.setting_mx_karma_key.get_value() is None:
			return

		if not self.api.activated:
			return

		current_map_length = int((datetime.datetime.now() - self.current_start).total_seconds())

		self.current_start = datetime.datetime.now()

		save_votes = []
		for vote in self.app.current_votes:
			login = vote.player.login
			score = vote.score
			if vote.expanded_score is not None:
				score = vote.expanded_score

			player_vote = []
			if self.current_votes is not None:
				player_vote = [v for v in self.current_votes if v['login'] == login]

			new_score = await self.determine_vote(score)
			if len(player_vote) == 0 or (len(player_vote) == 1 and player_vote[0]['vote'] != new_score):
				save_votes.append({'login': login, 'nickname': vote.player.nickname, 'vote': new_score})

		if len(save_votes) > 0:
			await self.api.save_votes(map=self.app.instance.map_manager.current_map, map_length=current_map_length, votes=save_votes)



from pyplanet.core import Controller
from pyplanet.core.events import Callback, handle_generic


async def handle_map_loading(source, signal, **kwargs):
	map = await Controller.instance.map_manager.get_map(uid=source['map']['uid'])
	return dict(
		map=map,
		restarted=source.get('restarted', None),
	)


server_start = Callback(
	call='Script.Maniaplanet.StartServer_Start',
	namespace='maniaplanet',
	code='server_start',
	target=handle_generic,
)
"""
:Signal: 
	Server Start signal
:Code:
	``maniaplanet:server_start``
:Description:
	This callback is called when the server script is (re)started. The begin of the event.
:Original Callback:
	`Script` Maniaplanet.StartServer_Start

:param restarted: Boolean giving information if the script has restarted.
:param time: Server time when callback has been sent.
"""


server_start__end = Callback(
	call='Script.Maniaplanet.StartServer_End',
	namespace='maniaplanet',
	code='server_start__end',
	target=handle_generic,
)
"""
:Signal: 
	Server Start signal (end of event).
:Code:
	``maniaplanet:server_start__end``
:Description:
	This callback is called when the server script is (re)started. The end of the event.
:Original Callback:
	`Script` Maniaplanet.StartServer_End

:param restarted: Boolean giving information if the script has restarted.
:param time: Server time when callback has been sent.
"""


server_end = Callback(
	call='Script.Maniaplanet.EndServer_Start',
	namespace='maniaplanet',
	code='server_end',
	target=handle_generic,
)
"""
:Signal: 
	Server End signal
:Code:
	``maniaplanet:server_end``
:Description:
	This callback is called when the server script is end. The begin of the event.
:Original Callback:
	`Script` Maniaplanet.EndServer_Start

:param restarted: Boolean giving information if the script has restarted.
:param time: Server time when callback has been sent.
"""


server_end__end = Callback(
	call='Script.Maniaplanet.EndServer_End',
	namespace='maniaplanet',
	code='server_end__end',
	target=handle_generic,
)
"""
:Signal: 
	Server End signal (end event)
:Code:
	``maniaplanet:server_end__end``
:Description:
	This callback is called when the server script is end. The end of the event.
:Original Callback:
	`Script` Maniaplanet.EndServer_End

:param restarted: Boolean giving information if the script has restarted.
:param time: Server time when callback has been sent.
"""


match_start = Callback(
	call='Script.Maniaplanet.StartMatch_Start',
	namespace='maniaplanet',
	code='match_start',
	target=handle_generic,
)
"""
:Signal: 
	Match Start.
:Code:
	``maniaplanet:match_start``
:Description:
	Callback sent when the "StartMatch" section start.
:Original Callback:
	`Script` Maniaplanet.StartMatch_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


match_start__end = Callback(
	call='Script.Maniaplanet.StartMatch_End',
	namespace='maniaplanet',
	code='match_start__end',
	target=handle_generic,
)
"""
:Signal: 
	Match Start. (End event)
:Code:
	``maniaplanet:match_start__end``
:Description:
	Callback sent when the "StartMatch" section end.
:Original Callback:
	`Script` Maniaplanet.StartMatch_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


match_end = Callback(
	call='Script.Maniaplanet.EndMatch_Start',
	namespace='maniaplanet',
	code='match_end',
	target=handle_generic,
)
"""
:Signal: 
	Match End.
:Code:
	``maniaplanet:match_end``
:Description:
	Callback sent when the "EndMatch" section start.
:Original Callback:
	`Script` Maniaplanet.EndMatch_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


match_end__end = Callback(
	call='Script.Maniaplanet.EndMatch_End',
	namespace='maniaplanet',
	code='match_end__end',
	target=handle_generic,
)
"""
:Signal: 
	Match End. (End event)
:Code:
	``maniaplanet:match_end__end``
:Description:
	Callback sent when the "EndMatch" section ends.
:Original Callback:
	`Script` Maniaplanet.EndMatch_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


round_start = Callback(
	call='Script.Maniaplanet.StartRound_Start',
	namespace='maniaplanet',
	code='round_start',
	target=handle_generic,
)
"""
:Signal: 
	Round Start.
:Code:
	``maniaplanet:round_start``
:Description:
	Callback sent when the "StartRound" section starts.
:Original Callback:
	`Script` Maniaplanet.StartRound_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


round_start__end = Callback(
	call='Script.Maniaplanet.StartRound_End',
	namespace='maniaplanet',
	code='round_start__end',
	target=handle_generic,
)
"""
:Signal: 
	Round Start. (End event)
:Code:
	``maniaplanet:round_start__end``
:Description:
	Callback sent when the "StartRound" section ends.
:Original Callback:
	`Script` Maniaplanet.StartRound_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


round_end = Callback(
	call='Script.Maniaplanet.EndRound_Start',
	namespace='maniaplanet',
	code='round_end',
	target=handle_generic,
)
"""
:Signal: 
	Round Start.
:Code:
	``maniaplanet:round_end``
:Description:
	Callback sent when the "EndRound" section starts.
:Original Callback:
	`Script` Maniaplanet.EndRound_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


round_end__end = Callback(
	call='Script.Maniaplanet.EndRound_End',
	namespace='maniaplanet',
	code='round_end__end',
	target=handle_generic,
)
"""
:Signal: 
	Round Start. (End event)
:Code:
	``maniaplanet:round_end__end``
:Description:
	Callback sent when the "EndRound" section ends.
:Original Callback:
	`Script` Maniaplanet.EndRound_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


turn_start = Callback(
	call='Script.Maniaplanet.StartTurn_Start',
	namespace='maniaplanet',
	code='turn_start',
	target=handle_generic,
)
"""
:Signal: 
	Turn Start.
:Code:
	``maniaplanet:turn_start``
:Description:
	Callback sent when the "StartTurn" section starts.
:Original Callback:
	`Script` Maniaplanet.StartTurn_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


turn_start__end = Callback(
	call='Script.Maniaplanet.StartTurn_End',
	namespace='maniaplanet',
	code='turn_start__end',
	target=handle_generic,
)
"""
:Signal: 
	Turn Start. (End event).
:Code:
	``maniaplanet:turn_start__end``
:Description:
	Callback sent when the "StartTurn" section ends.
:Original Callback:
	`Script` Maniaplanet.StartTurn_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


turn_end = Callback(
	call='Script.Maniaplanet.EndTurn_Start',
	namespace='maniaplanet',
	code='turn_end',
	target=handle_generic,
)
"""
:Signal: 
	Turn End.
:Code:
	``maniaplanet:turn_end``
:Description:
	Callback sent when the "EndTurn" section starts.
:Original Callback:
	`Script` Maniaplanet.EndTurn_Start

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


turn_end__end = Callback(
	call='Script.Maniaplanet.EndTurn_End',
	namespace='maniaplanet',
	code='turn_end__end',
	target=handle_generic,
)
"""
:Signal: 
	Turn End. (End event)
:Code:
	``maniaplanet:turn_end__end``
:Description:
	Callback sent when the "EndTurn" section ends.
:Original Callback:
	`Script` Maniaplanet.EndTurn_End

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


play_loop_start = Callback(
	call='Script.Maniaplanet.StartPlayLoop',
	namespace='maniaplanet',
	code='play_loop_start',
	target=handle_generic,
)
"""
:Signal: 
	Play Loop Start.
:Code:
	``maniaplanet:play_loop_start``
:Description:
	Callback sent when the "PlayLoop" section starts.
:Original Callback:
	`Script` Maniaplanet.StartPlayLoop

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


play_loop_end = Callback(
	call='Script.Maniaplanet.EndPlayLoop',
	namespace='maniaplanet',
	code='play_loop_end',
	target=handle_generic,
)
"""
:Signal: 
	Play Loop End.
:Code:
	``maniaplanet:play_loop_end``
:Description:
	Callback sent when the "PlayLoop" section ends.
:Original Callback:
	`Script` Maniaplanet.EndPlayLoop

:param count: Each time this section is played, this number is incremented by one.
:param time: Server time when callback has been sent.
"""


loading_map_start = Callback(
	call='Script.Maniaplanet.LoadingMap_Start',
	namespace='maniaplanet',
	code='loading_map_start',
	target=handle_generic
)
"""
:Signal: 
	Loading Map start.
:Code:
	``maniaplanet:loading_map_start``
:Description:
	Callback sent when the server starts loading the map.
:Original Callback:
	`Script` Maniaplanet.LoadingMap_Start

:param time: Server time when callback has been sent.
"""


loading_map_end = Callback(
	call='Script.Maniaplanet.LoadingMap_End',
	namespace='maniaplanet',
	code='loading_map_end',
	target=handle_map_loading
)
"""
:Signal: 
	Loading Map end.
:Code:
	``maniaplanet:loading_map_end``
:Description:
	Callback sent when the server finishes to load the map.
:Original Callback:
	`Script` Maniaplanet.LoadingMap_End

:param map: Map instance from database. Updated with the provided data.
:type map: pyplanet.core.maniaplanet.models.map.Map
"""


unloading_map_start = Callback(
	call='Script.Maniaplanet.UnloadingMap_Start',
	namespace='maniaplanet',
	code='unloading_map_start',
	target=handle_map_loading
)
"""
:Signal: 
	Unloading of the Map starts.
:Code:
	``maniaplanet:unloading_map_start``
:Description:
	Callback sent when the server starts to unload a map.
:Original Callback:
	`Script` Maniaplanet.UnloadingMap_Start

:param map: Map instance from database. Updated with the provided data.
:type map: pyplanet.core.maniaplanet.models.map.Map
"""


unloading_map_end = Callback(
	call='Script.Maniaplanet.UnloadingMap_End',
	namespace='maniaplanet',
	code='unloading_map_end',
	target=handle_generic
)
"""
:Signal: 
	Unloading of the Map ends.
:Code:
	``maniaplanet:unloading_map_end``
:Description:
	Callback sent when the server finishes to unload a map.
:Original Callback:
	`Script` Maniaplanet.UnloadingMap_End

:param map: Map instance from database. Updated with the provided data.
:type map: pyplanet.core.maniaplanet.models.map.Map
"""


podium_start = Callback(
	call='Script.Maniaplanet.Podium_Start',
	namespace='maniaplanet',
	code='podium_start',
	target=handle_generic
)
"""
:Signal: 
	Podium start.
:Code:
	``maniaplanet:podium_start``
:Description:
	Callback sent when the podium sequence starts.
:Original Callback:
	`Script` Maniaplanet.Podium_Start

:param time: Server time when callback has been sent.
"""


podium_end = Callback(
	call='Script.Maniaplanet.Podium_End',
	namespace='maniaplanet',
	code='podium_end',
	target=handle_generic
)
"""
:Signal: 
	Podium end.
:Code:
	``maniaplanet:podium_end``
:Description:
	Callback sent when the podium sequence ends.
:Original Callback:
	`Script` Maniaplanet.Podium_End

:param time: Server time when callback has been sent.
"""


match_begin = Callback(
	call='ManiaPlanet.BeginMatch',
	namespace='maniaplanet',
	code='match_begin',
	target=handle_generic,
)
match_stop = Callback(
	call='ManiaPlanet.EndMatch',
	namespace='maniaplanet',
	code='match_stop',
	target=handle_generic,
)


status_changed = Callback(
	call='ManiaPlanet.StatusChanged',
	namespace='maniaplanet',
	code='status_changed',
	target=handle_generic,
)
"""
:Signal: 
	Server Status Changed.
:Code:
	``maniaplanet:status_changed``
:Description:
	Callback sent when the podium sequence ends.
:Original Callback:
	`Native` Maniaplanet.Podium_End

:param 1: Status Code.
:param 2: Status Name.
:type 1: int
:type 2: str
"""



from .base import StatsView


class StatsDashboardView(StatsView):
	"""
	The Statistics Dashboard will show the summary of some statistics about the player and the server. The player can click
	on one of the 'More' buttons to show more detailed statistics about the subject.
	"""
	template_name = 'core.statistics/dashboard.xml'

	def __init__(self, app, *args, **kwargs):
		"""
		Init the dashboard.

		:param app: App instance
		:param args: Args..
		:param kwargs: Kwargs...
		:type app: pyplanet.apps.core.statistics.Statistics
		"""
		super().__init__(*args, **kwargs)
		self.app = app

		self.subscribe('button_close', self.close)

	async def get_context_data(self):
		context = await super().get_context_data()
		context['options'] = [
			dict(name='Top Records', view='pyplanet.apps.core.statistics.views.records.TopRecordsView'),
		]
		context['sizes'] = {
			'top__pos': '0 67',
			'top__size': '206.5 8',
			'box__size': '200 125',
			'bottom__pos': '0 -62',
			'bottom__size': '200 2',
			'input__pos': '0 -40',
			'input__size': '185 7',
		}
		return context

	async def get_all_player_data(self, logins):
		data = dict()
		for login in logins:
			player = await self.app.instance.player_manager.get_player(login=login, lock=False)
			data[login] = await self.app.processor.get_dashboard_data(player)
			# data[login]['numbers_chart'] = 'https://image-charts.com/chart?cht=p&chd=t:10,90&chs=200x200&chl=Top 3%7CRecords%7C.png'

		return data

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])
		await self.destroy()



from pyplanet.apps.contrib.karma.models import Karma
from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.views.generics.widget import WidgetView
from pyplanet.views.generics.list import ManualListView


class KarmaWidget(WidgetView):
	widget_x = 125
	widget_y = 70
	z_index = 130

	template_name = 'karma/karma.xml'

	def __init__(self, app):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.id = 'pyplanet__widgets_karma'

		self.action = self.action_whokarma
		self.subscribe('vote_positive', self.action_vote_positive)
		self.subscribe('vote_negative', self.action_vote_negative)

	async def get_context_data(self):
		context = await super().get_context_data()

		karma_percentage = round(self.app.current_karma_percentage * 100, 1) \
			if len(self.app.current_votes) > 0 else 0
		bar_width = round(((karma_percentage / 100) * 24), 2)

		context.update({
			'current_karma': self.app.current_karma,
			'karma_percentage': karma_percentage,
			'progress_percentage': karma_percentage if karma_percentage != 0 else 5,
			'positive_votes': round(self.app.current_karma_positive, 2),
			'negative_votes': round(self.app.current_karma_negative, 2),
			'bar_width': bar_width,
		})

		return context

	async def get_player_data(self):
		data = await super().get_player_data()
		votes = dict()

		for player in self.app.instance.player_manager.online:
			player_vote = [x for x in self.app.current_votes if x.player_id == player.get_id()]
			if len(player_vote) > 0:
				if player_vote[0].expanded_score is not None:
					votes[player.login] = {'player_vote': player_vote[0].expanded_score}
				else:
					votes[player.login] = {'player_vote': player_vote[0].score}
			else:
				votes[player.login] = {'player_vote': 0}

		data.update(votes)

		return data

	async def action_whokarma(self, player, **kwargs):
		await self.app.show_map_list(player)

	async def action_vote_positive(self, player, action, values, **kwargs):
		await self.app.player_chat(player, '++', False)

	async def action_vote_negative(self, player, action, values, **kwargs):
		await self.app.player_chat(player, '--', False)


class KarmaListView(ManualListView):
	title = 'Karma votes on this map'
	icon_style = 'Icons128x128_1'
	icon_substyle = 'Buddies'

	fields = [
		{
			'name': 'Player',
			'index': 'nickname',
			'sorting': False,
			'searching': True,
			'width': 70
		},
		{
			'name': 'Vote',
			'index': 'vote',
			'sorting': True,
			'searching': False,
			'width': 30,
			'type': 'label'
		}
	]

	def __init__(self, app, map):
		super().__init__(self)
		self.app = app
		self.manager = app.context.ui
		self.map = map

	async def get_title(self):
		return 'Karma votes on {}'.format(self.map.name)

	async def get_data(self):
		karma = await Karma.execute(
			Karma
				.select(Karma, Player)
				.join(Player)
				.where(Karma.map_id == self.map.id)
		)

		votes = []
		for item in karma:
			score = item.score
			if item.expanded_score is not None:
				score = item.expanded_score

			vote = '+-'
			if score == 1:
				vote = '++'
			elif score == 0.5:
				vote = '+'
			elif score == -0.5:
				vote = '-'
			elif score == -1:
				vote = '--'

			votes.append({'nickname': item.player.nickname, 'vote': vote})

		return votes



#!/usr/bin/env python3
"""
Simulate the CLI interface without any project.
"""
import os

if __name__ == '__main__':
	from pyplanet.core.management import execute_from_command_line
	execute_from_command_line()



from pyplanet.views import TemplateView


class ToolbarView(TemplateView):
	template_name = 'core.pyplanet/toolbar.xml'

	def __init__(self, app, *args, **kwargs):
		"""
		Initiate Player Toolbar

		:param app: App instance.
		:type app: pyplanet.apps.core.pyplanet.app.PyPlanetConfig
		"""
		super().__init__(*args, **kwargs)

		self.id = 'pyplanet__toolbar'
		self.app = app
		self.manager = self.app.context.ui

		self.commands = {
			'bar_button_list': '/list',
			'bar_button_mf': '/mf',
			'bar_button_jb': '/jukebox list',
			'bar_button_skip': '/skip',
			'bar_button_extend': '/extend',
			'bar_button_replay': '/replay',
			'bar_button_players': '/players',
			'bar_button_topdons': '/topdons',
			'bar_button_topsums': '/topsums',
			'bar_button_topactive': '/topactive',
			'bar_button_mxinfo': '/{} info'.format('tmx' if self.app.instance.game.game == 'tmnext' else 'mx'),
			'bar_button_help': '/helpall',
		}

	async def get_context_data(self):
		data = await super().get_context_data()
		data['game'] = self.app.instance.game.game
		return data

	async def handle_catch_all(self, player, action, values, **kwargs):
		if action not in self.commands:
			return
		await self.app.instance.command_manager.execute(player, self.commands[action])



import asynctest

from pyplanet.apps import AppConfig
from pyplanet.core import Controller
from pyplanet.core.events import Signal


class TestSignals(asynctest.TestCase):
	async def test_registering(self):
		instance = Controller.prepare(name='default').instance

		test1 = Signal(code='test1', namespace='tests')
		instance.signals.register_signal(test1)
		test2 = Signal(code='test2', namespace='tests')
		instance.signals.register_signal(test2)

		test1_comp = instance.signals.get_signal('tests:test1')
		test2_comp = instance.signals.get_signal('tests:test2')

		assert test1 == test1_comp
		assert test2 == test2_comp

	async def test_listening(self):
		instance = Controller.prepare(name='default').instance

		test1 = Signal(code='test1', namespace='tests', process_target=self.glue)
		instance.signals.register_signal(test1)

		self.got_sync = 0
		self.got_async = 0
		self.got_glue = 0
		self.got_raw = 0

		test1.register(self.sync_listener)
		test1.register(self.async_listener)

		await test1.send(dict(glue=False), raw=True)
		await test1.send(dict(glue=False), raw=False)
		await test1.send_robust(dict(glue=False), raw=True)
		await test1.send_robust(dict(glue=False), raw=False)

		assert self.got_async == 4
		assert self.got_sync == 4
		assert self.got_glue == 2
		assert self.got_raw == 4

	async def test_unregister(self):
		instance = Controller.prepare(name='default').instance
		test1 = Signal(code='test1', namespace='tests', process_target=self.glue)
		instance.signals.register_signal(test1)

		self.got_sync = 0
		self.got_async = 0
		self.got_glue = 0
		self.got_raw = 0

		test1.register(self.sync_listener)
		test1.register(self.async_listener)

		await test1.send(dict(glue=False))

		# Unregister + send
		test1.unregister(self.sync_listener)

		await test1.send(dict(glue=False), raw=True)

		assert self.got_sync == 1
		assert self.got_async == 2

	async def test_app_manager(self):
		instance = Controller.prepare(name='default').instance
		manager = instance.signals.create_app_manager(None)

		test1 = Signal(code='test1', namespace='tests', process_target=self.glue)
		manager.register_signal(test1)

		self.got_sync = 0
		self.got_async = 0
		self.got_glue = 0
		self.got_raw = 0

		manager.listen(test1, self.sync_listener)
		manager.listen(test1, self.async_listener)

		await test1.send(dict(glue=False))

		await manager.on_destroy()

		# This one will be ignored because the app is destroyed.
		await test1.send(dict(glue=False), raw=True)

		assert self.got_sync == 1
		assert self.got_async == 1

	####################################################################################################################

	def sync_listener(self, glue=None, source=None, **kwargs):
		self.got_sync += 1
		if glue is not True:
			self.got_raw += 1
		if source:
			assert isinstance(source, dict)

	async def async_listener(self, glue=None, source=None, **kwargs):
		self.got_async += 1
		if glue is not True:
			self.got_raw += 1
		if source:
			assert isinstance(source, dict)

	async def glue(self, source, signal):
		self.got_glue += 1
		return dict(glue=True)



"""
The MX Karma API client class.
"""
import logging
import aiohttp
import hashlib

from pyplanet import __version__ as version
from pyplanet.apps.contrib.mx.exceptions import MXInvalidResponse

logger = logging.getLogger(__name__)


class MXKarmaApi:
	"""
	MX Karma API Class will handle all the communication in async with the ManiaExchange Karma API.
	"""
	def __init__(self, app):
		self.api_url = 'https://karma.mania-exchange.com/api2'
		self.app = app
		self.cookie_jar = aiohttp.CookieJar()
		self.session = None
		self.key = None
		self.activated = False

	async def create_session(self):
		"""
		Create new async http session.
		"""
		if self.session is None:
			logger.debug('Creating session for MX Karma communication.')

			self.session = await aiohttp.ClientSession(
				cookie_jar=self.cookie_jar,
				headers={
					'User-Agent': 'PyPlanet/{}'.format(version),
					'X-ManiaPlanet-ServerLogin': self.app.app.instance.game.server_player_login
				}
			).__aenter__()

	async def close_session(self):
		"""
		Close the async http session
		"""
		if self.session and hasattr(self.session, '__aexit__'):
			logger.debug('Closing session for MX Karma communication.')

			await self.session.__aexit__(None, None, None)

	async def start_session(self):
		"""
		Create the MX Karma session by authenticating with the API. This method will make sure the server is ready
		to get and post karma entries to the API.
		"""
		if self.session is None:
			await self.create_session()
		elif self.session is not None and self.key is None:
			try:
				logger.debug('Starting MX Karma session ...')

				url = '{url}/startSession?serverLogin={login}&applicationIdentifier={app}&testMode=false'.format(
					url=self.api_url, login=self.app.app.instance.game.server_player_login,
					app='PyPlanet {version}'.format(version=version)
				)

				response = await self.session.get(url)

				if response.status < 200 or response.status > 399:
					raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

				result = await response.json()
				if result['success'] is False:
					logger.error('Error while starting ManiaExchange session, error {code}: {message}'.format(
						code=result['data']['code'], message=result['data']['message']
					))
				else:
					self.key = result['data']['sessionKey']
					await self.activate_session(result['data']['sessionSeed'])
			except (aiohttp.ClientConnectionError, aiohttp.ClientConnectorCertificateError):
				logger.warning('Unable to start a MX Karma session, server is unavailable.')
			except MXInvalidResponse:
				logger.warning('Unable to start a MX Karma session, invalid server response.')

	async def activate_session(self, session_seed):
		"""
		This method will activate the session based on the seed it created earlier.

		:param session_seed: Seed of the session created earlier.
		"""
		api_key = await self.app.setting_mx_karma_key.get_value()
		hash = hashlib.sha512('{apikey}{seed}'.format(apikey=api_key, seed=session_seed).encode('utf-8')).hexdigest()

		url = '{url}/activateSession?sessionKey={key}&activationHash={hash}'.format(
			url=self.api_url, key=self.key, hash=hash
		)

		response = await self.session.get(url)

		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

		result = await response.json()

		if result['success'] is False:
			self.activated = False
			logger.error('Error while getting information from ManiaExchange, error {code}: {message}'.format(
				code=result['data']['code'], message=result['data']['message']
			))
		else:
			self.activated = result['data']['activated']

			if self.activated:
				logger.info('Successfully started the MX Karma session.')
			else:
				logger.warning('Unable to start a MX Karma session.')

	async def get_map_rating(self, map, player_login = None):
		"""
		Get rating for the specific map instance.

		:param map: Map model instance.
		:param player_login: Player login, optional
		:return: Dictionary with data for the given map, received from the API.
		"""
		if not self.activated:
			return

		url = '{url}/getMapRating?sessionKey={key}'.format(
			url=self.api_url, key=self.key
		)

		logins = [vote.player.login for vote in self.app.app.current_votes]
		loop_logins = [logins[x:x+100] for x in range(0, len(logins), 100)]

		data = None

		for temp_logins in loop_logins:
			content = {
				'gamemode': await self.app.app.instance.mode_manager.get_current_full_script(),
				'titleid': self.app.app.instance.game.dedicated_title,
				'mapuid': map.uid,
				'getvotesonly': False,
				'playerlogins': temp_logins
			}

			response = await self.session.post(url, json=content)

			if response.status < 200 or response.status > 399:
				raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

			result = await response.json()

			if result['success'] is False:
				logger.error('Error while getting information from ManiaExchange, error {code}: {message}'.format(
					code=result['data']['code'], message=result['data']['message']
				))

				return None
			else:
				if data is None:
					data = result['data']
				else:
					data['votes'] += result['data']['votes']

		if data and 'votes' in data:
			logger.debug('Amount of MX Karma votes retrieved for this map: {}'.format(len(data['votes'])))

		return data

	async def save_votes(self, map, is_import=False, map_length=0, votes=None):
		"""
		Save votes to the MX Karma API.

		:param map: Map instance
		:param is_import: State if the votes are imported from local karma.
		:param map_length: Map length.
		:param votes: Votes payload.
		:return: Status or updated value.
		"""
		if not self.activated:
			return

		if votes is None:
			return

		url = '{url}/saveVotes?sessionKey={key}'.format(
			url=self.api_url, key=self.key
		)

		content = {
			'gamemode': await self.app.app.instance.mode_manager.get_current_full_script(),
			'titleid': self.app.app.instance.game.dedicated_title,
			'mapuid': map.uid,
			'mapname': map.name,
			'mapauthor': map.author_login,
			'isimport': is_import,
			'maptime': map_length,
			'votes': votes
		}

		response = await self.session.post(url, json=content)

		if response.status < 200 or response.status > 399:
			raise MXInvalidResponse('Got invalid response status from ManiaExchange: {}'.format(response.status))

		result = await response.json()

		if result['success'] is False:
			logger.error('Error while saving information to ManiaExchange, error {code}: {message}'.format(
				code=result['data']['code'], message=result['data']['message']
			))
			return False
		else:
			if votes is not None:
				logger.debug('Amount of MX Karma votes saved for this map: {}'.format(len(votes)))
			return result['data']['updated']



"""
The chat contrib makes it possible to send chat messages way more easy and faster. It also maintains some other features
related to the chat.
"""
from .manager import ChatManager

__all__ = [
	'ChatManager'
]



from pyplanet.contrib.setting import Setting

# Performance mode setting
performance_mode = Setting(
	'performance_mode', 'Performance mode playercount', Setting.CAT_BEHAVIOUR, type=int,
	description='Above this amount of players the performance mode will be enabled.',
	default=30
)



"""
Toolbar View.
"""
from pyplanet.views import TemplateView


class EmojiToolbarView(TemplateView):
	template_name = 'funcmd/emoji_toolbar.xml'

	def __init__(self, app):
		"""
		Initiate the toolbar view.

		:param app: Player app instance.
		:type app: pyplanet.apps.contrib.funcmd.FunCmd
		"""
		super().__init__(app.context.ui)

		self.app = app
		self.instance = app.instance
		self.id = 'funcmd_emoji_toolbar'
		self.emoji_list = [
			'',  # Smiley
			'',  # Neutral Smiley
			'',  # Unhappy Smiley
			'$F30$z',  # Thumbs-down
			'$393$z',  # Thumbs-up
			'',  # Two fingers left
			'',  # Two fingers right
			'$F30$z',  # Heart <3
			'',  # Map/locator
			'$cc0🏆$z',  # Cup
		]

		for i in range(1, 11):
			self.subscribe('bar_button_{}'.format(i), self.action_emoji(self.emoji_list[i-1]))

	async def get_context_data(self):
		data = await super().get_context_data()
		data['game'] = self.app.instance.game.game
		return data

	def action_emoji(self, emoji):
		async def send_emoji(player, *args, **kwargs):
			if 'admin' in self.instance.apps.apps and self.instance.apps.apps['admin'].server.chat_redirection:
				return
			await self.instance.chat(
				'$z[{}$z] $z{}'.format(player.nickname, emoji)
			)
		return send_emoji



import logging

from asyncio import iscoroutinefunction

from pyplanet.apps.core.maniaplanet.models import Player
from pyplanet.views.template import TemplateView

logger = logging.getLogger(__name__)


class WidgetView(TemplateView):
	widget_x = None
	widget_y = None
	size_x = None
	size_y = None
	title = None
	action = None
	z_index = None
	distraction_hide = True

	template_name = 'core.views/generics/widget.xml'

	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		self.relaxed_updating = True

		# Setup the receivers.
		self.subscribe('open_widget', self.open_widget)

	async def open_widget(self, player, action, values, **kwargs):
		if self.action is not None:
			# Execute action/target method.
			if iscoroutinefunction(self.action):
				await self.action(player)
			else:
				self.action(player)

	async def close(self, player, *args, **kwargs):
		"""
		Close the link for a specific player. Will hide manialink and destroy data for player specific to save memory.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		if self.player_data and player.login in self.player_data:
			del self.player_data[player.login]
		await self.hide(player_logins=[player.login])

	async def refresh(self,  player=None, *args, **kwargs):
		"""
		Refresh list with current properties for a specific player. Can be used to show new data changes.

		:param player: Player model instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		"""
		await self.display(player=player)

	async def display(self, player=None, **kwargs):
		"""
		Display list to player.

		:param player: Player login or model instance.
		:type player: str, pyplanet.apps.core.maniaplanet.models.Player
		"""
		login = player.login if isinstance(player, Player) else player
		if not player:
			return await super().display(**kwargs)
		return await super().display(player_logins=[login], **kwargs)

	async def get_title(self):
		return self.title

	async def get_context_data(self):
		context = await super().get_context_data()

		icon_x = 0.5
		if self.widget_x < 0 and self.size_x:
			icon_x = (self.size_x - 2.5 - 4.5)

		hover_color = '3341'
		if self.action is not None:
			hover_color = '09F4'

		# Add facts.
		context.update({
			'widget_x': self.widget_x,
			'widget_y': self.widget_y,
			'z_index': self.z_index,
			'distraction_hide': self.distraction_hide,
			'size_x': self.size_x,
			'size_y': self.size_y,
			'hover_color': hover_color,
			'icon_x': icon_x,
			'title': await self.get_title(),
			'open_action': self.action is not None,
			'content_pos_x': 2,
			'content_pos_y': -5,
		})

		return context


class TimesWidgetView(WidgetView):
	template_name = 'core.views/generics/timeswidget.xml'

	async def get_context_data(self):
		context = await super().get_context_data()

		# Add facts.
		context.update({
			'content_pos_x': 1,
			'content_pos_y': -4.5,
			'top_entries': 3,
			'times': None
		})

		return context



from pyplanet.core.events import Callback, handle_generic
from pyplanet.core.instance import Controller


async def handle_map_begin(source, signal, **kwargs):
	# Retrieve and update map, return it to our callback listeners.
	map = await Controller.instance.map_manager.handle_map_change(source)
	return dict(map=map)

async def handle_map_end(source, signal, **kwargs):
	# Retrieve and update map, return it to our callback listeners.
	map = await Controller.instance.map_manager.handle_map_change(source)
	return dict(map=map)

async def handle_playlist_modified(source, signal, **kwargs):
	# Make sure the map manager fetches the updated maplist.
	updated = await Controller.instance.map_manager.handle_playlist_change(source)
	if not isinstance(source, dict):
		source = dict()
	source['maps_updated'] = updated
	return source

async def handle_map_start(source, signal, **kwargs):
	# Get the map
	map = await Controller.instance.map_manager.get_map(source['map']['uid'])
	return dict(
		time=source['time'], count=source['count'], restarted=source['restarted'], map=map,
	)


# We don't use scripted map events due to the information we get and the stability of the information structure.
map_begin = Callback(
	call='ManiaPlanet.BeginMap',
	namespace='maniaplanet',
	code='map_begin',
	target=handle_map_begin,
)
"""
:Signal: 
	Begin of map.
:Code:
	``maniaplanet:map_begin``
:Description:
	Callback sent when map begins.
:Original Callback:
	`Native` Maniaplanet.BeginMap

:param map: Map instance.
:type map: pyplanet.apps.core.maniaplanet.models.map.Map
"""

map_end = Callback(
	call='ManiaPlanet.EndMap',
	namespace='maniaplanet',
	code='map_end',
	target=handle_map_end,
)
"""
:Signal: 
	End of map.
:Code:
	``maniaplanet:map_end``
:Description:
	Callback sent when map ends.
:Original Callback:
	`Native` Maniaplanet.EndMap

:param map: Map instance.
:type map: pyplanet.apps.core.maniaplanet.models.map.Map
"""

playlist_modified = Callback(
	call='ManiaPlanet.MapListModified',
	namespace='maniaplanet',
	code='playlist_modified',
	target=handle_playlist_modified
)
"""
:Signal: 
	Maplist changes.
:Code:
	``maniaplanet:playlist_modified``
:Description:
	Callback sent when map list changes.
:Original Callback:
	`Native` Maniaplanet.MapListModified

:param 1: Current map index.
:param 2: Next map index.
:param 3: Is List Modified.
:type 1: int
:type 2: int
:type 3: bool
"""

map_start = Callback(
	call='Script.Maniaplanet.StartMap_Start',
	namespace='maniaplanet',
	code='map_start',
	target=handle_map_start
)
"""
:Signal: 
	Begin of map. (Scripted!)
:Code:
	``maniaplanet:map_begin``
:Description:
	Callback sent when map starts (same as begin, but scripted).
:Original Callback:
	`Script` Maniaplanet.StartMap_Start

:param time: Time when callback has been sent.
:param count: Counts of the callback that was sent.
:param restarted: Is the map restarted.
:param map: Map instance.
:type map: pyplanet.apps.core.maniaplanet.models.map.Map
"""

map_start__end = Callback(
	call='Script.Maniaplanet.StartMap_End',
	namespace='maniaplanet',
	code='map_start__end',
	target=handle_map_start
)
"""
:Signal: 
	Begin of map, end of event. (Scripted!)
:Code:
	``maniaplanet:map_start__end``
:Description:
	Callback sent when map starts (same as begin, but scripted). End of the event
:Original Callback:
	`Script` Maniaplanet.StartMap_End

:param time: Time when callback has been sent.
:param count: Counts of the callback that was sent.
:param restarted: Is the map restarted.
:param map: Map instance.
:type map: pyplanet.apps.core.maniaplanet.models.map.Map
"""




"""
This file contains the default configuration for PyPlanet. All configuration the user provides will override the
following lines declared in this file.
"""
import tempfile
import os
import logging


##########################################
################# CORE ###################
##########################################

# Enable debug mode to get verbose output, not report any errors and dynamically use the DEBUG in your code
# for extra verbosity of logging/output.
DEBUG = False

# This should ALWAYS be overridden by the local settings.
ROOT_PATH = None

# Define the temporary folder to write temporary files to, such as downloaded files that are only required once
# or are only required parsing and can be removed.
# This should ALWAYS be overridden by the local settings.
TMP_PATH = None

# Add your pools (the controller instances per dedicated here) or leave as it is to use a single instance only.
POOLS = [
	'default'
]

# Allow self-upgrading the installation. Disable on shared servers with one installation!
SELF_UPGRADE = True

##########################################
################## DB ####################
##########################################

# Databases configuration holds an dictionary with information of the database backend.
# Please refer to the documentation for all examples.
DATABASES = {
	'default': {
		'ENGINE': 'peewee.SqliteDatabase',
		'NAME': 'database.db',
	}
}


##########################################
################ CACHE ###################
##########################################

# Define any cache backends that can be used by the core and the plugins to cache data.
CACHES = {
	'default': {
		'DRIVER': 'pyplanet.cache.backends.memory',
	}
}


##########################################
############### LOGGING ##################
##########################################

# Logging configuration handler. Defaults to dictionary configuration.
LOGGING_CONFIG = 'logging.config.dictConfig'

# Logging configuration.
LOGGING = {
	'version': 1,
	'disable_existing_loggers': False,
	'filters': {
		'require_debug_false': {
			'()': 'pyplanet.utils.log.RequireDebugFalse',
		},
		'require_debug_true': {
			'()': 'pyplanet.utils.log.RequireDebugTrue',
		},
		'require_exception': {
			'()': 'pyplanet.utils.log.RequireException',
		}
	},
	'formatters': {
		'colored': {
			'()': 'colorlog.ColoredFormatter',
			'format': "%(log_color)s%(levelname)-8s%(reset)s %(yellow)s[%(threadName)s][%(name)s]%(reset)s %(white)s%(message)s",
		},
		'timestamped': {
			'format': '[%(asctime)s][%(levelname)s][%(threadName)s] %(name)s: %(message)s (%(filename)s:%(lineno)d)',
		},
	},
	'handlers': {
		'console-debug': {
			'class': 'logging.StreamHandler',
			'filters': ['require_debug_true'],
			'formatter': 'colored',
			'level': logging.DEBUG,
		},
		'console': {
			'class': 'logging.StreamHandler',
			'filters': ['require_debug_false'],
			'formatter': 'colored',
			'level': logging.INFO,
		}
	},
	'loggers': {
		'pyplanet': {
			'handlers': ['console', 'console-debug'],
			'level': logging.DEBUG,
			'propagate': False,
		}
	},
	'root': {
		'handlers': ['console', 'console-debug'],
		'level': logging.DEBUG,
	}
}

# Global logging handlers.
LOGGING_WRITE_LOGS = False
LOGGING_ROTATE_LOGS = True
LOGGING_DIRECTORY = 'logs'

# Error reporting
# See documentation for the options, (docs => privacy).
# Options:
# 0 = Don't report any errors or messages.
# 1 = Report errors with only traces.
# 2 = Report errors with traces and server data.
# 3 = Report errors with traces and server data, provide data to contributed apps (only pyplanet team has access).
LOGGING_REPORTING = 3

# Enable usage analytics. On by default. (Will be turned off when DEBUG is true!).
ANALYTICS = True


##########################################
################# APPS ###################
##########################################
APPS = {
	'default': []
}

# The following apps are mandatory loaded, and part of the core. This apps are always loaded *BEFORE* all other
# apps are initiated and loaded.
MANDATORY_APPS = [
	'pyplanet.apps.core.pyplanet.app.PyPlanetConfig',
	'pyplanet.apps.core.maniaplanet.app.ManiaplanetConfig',
	'pyplanet.apps.core.trackmania.app.TrackmaniaConfig',
	'pyplanet.apps.core.shootmania.app.ShootmaniaConfig',
	'pyplanet.apps.core.statistics.app.StatisticsConfig',
]

##########################################
############## DEDICATED #################
##########################################

# Dedicated contains the dedicated servers configurations, by default this is the localhost entry with default
# credentials and details.
DEDICATED = {
	'default': {
		'HOST': '127.0.0.1',
		'PORT': '5000',
		'USER': 'SuperAdmin',
		'PASSWORD': 'SuperAdmin',
	}
}

# Map configuration is a set of configuration options related to match settings etc.
# Matchsettings filename.
MAP_MATCHSETTINGS = {
	'default': 'maplist.txt'
}

# Blacklist file is managed by the dedicated server and will be loaded and writen to by PyPlanet once a
# player gets blacklisted. The default will be the filename Maniaplanet always uses and is generic.
BLACKLIST_FILE = {
	'default': 'blacklist.txt'
}

# The storage configuration contains the same instance mapping of the dedicated servers and is used
# to access the filesystem on the dedicated server location.
# Please refer to the documentation for more information.
STORAGE = {
	'default': {
		'DRIVER': 'pyplanet.core.storage.drivers.local.LocalDriver',
		'OPTIONS': {},
	}
}

# Owners are logins of the server owners, the owners always get *ALL* the permissions in the system.
OWNERS = {
	'default': []
}

# Songs is a list of URLs to .ogg files of songs to be played by the music server.
SONGS = {
	'default': []
}




import asyncio
from inspect import iscoroutinefunction

from pyplanet.contrib.command.params import ParameterParser


class Command:
	"""
	The command instance describes the command itself, the target to fire and all other related information, like
	admin command or aliases.

	Some examples of some commands:

	.. code-block:: python

		# Admin command with permission on it.
		Command(command='reboot', target=self.reboot_pool, perms='admin:reboot', admin=True)

		# Normal user command with optional argument.
		Command(command='list', target=self.show_map_list)\
			.add_param(name='search', required=False)

	"""

	def __init__(
		self, command, target, aliases=None, admin=False, namespace=None, parser=None, perms=None, description=None
	):
		"""
		Initiate a command.

		:param command: Command text (prefix without parameters).
		:param target: Target method to fire.
		:param aliases: Alias(ses) for the command.
		:param admin: Register command in admin context.
		:param namespace: Custom namespace, this can be used to create commands like '/prog start' and '/prog end'
						  where 'prog' is the namespace.
		:param perms: Required parameters, default everyone is allowed.
		:param parser: Custom parser.
		:param description: Description of the command.
		:type command: str
		:type target: any
		:type aliases: str[]
		:type admin: bool
		:type namespace: str, str[]
		:type perms: list,str
		:type parser: any
		:type description: str
		"""
		self.command = command
		self.target = target
		self.aliases = aliases or list()
		self.admin = admin
		self.namespace = namespace
		if isinstance(perms, str):
			perms = [perms]
		self.perms = perms
		self.parser = parser or \
					  ParameterParser('{} {}'.format(self.namespace, self.command) if self.namespace else self.command)
		self.description = description

	def match(self, raw):
		"""
		Try to match the command with the given input in array style (splitted by spaces).

		:param raw: Raw input, split by spaces.
		:type raw: list
		:return: Boolean if command matches.
		"""
		input = raw[:]

		if len(input) == 0 or (len(input) == 1 and input[0] == ''):
			return False

		if self.admin:
			if input[0][0:1] == '/':
				input[0] = input[0][1:]
			elif input[0] == 'admin':
				input.pop(0)
			else:
				return False

		# Make sure namespace is always an array if provided.
		if self.namespace and not isinstance(self.namespace, (list, tuple)):
			self.namespace = [self.namespace]

		# Check against namespace.
		if len(input) > 0 and self.namespace and any(input[0] == n for n in self.namespace):
			input.pop(0)
		elif self.namespace:
			return False

		if not len(input):
			return False

		command = input.pop(0)
		if self.command == command or command in self.aliases:
			return True
		return False

	def get_params(self, input):
		"""
		Get params in array from input in array.

		:param input: Array of raw input.
		:type input: list
		:return: Array of parameters, stripped of the command name and namespace, if defined.
		:rtype: list
		"""
		if self.admin:
			if input[0][0:1] == '/':
				input[0] = input[0][1:]
			elif input[0] == 'admin':
				input.pop(0)
		if self.namespace:
			input.pop(0)
		input.pop(0)
		return input

	def add_param(
		self, name: str,
		nargs=1,
		type=str,
		default=None,
		required: bool=True,
		help: str=None,
		dest: str=None,
	):
		"""
		Add positional parameter.

		:param name: Name of parameter, will be used to store result into!
		:param nargs: Number of arguments, use integer or '*' for multiple or infinite.
		:param type: Type of value, keep str to match all types. Use any other to try to parse to the type.
		:param default: Default value when no value is given.
		:param required: Set the parameter required state, defaults to true.
		:param help: Help text to display when parameter is invalid or not given and required.
		:param dest: Destination to save into namespace result (defaults to name).
		:return: parser instance
		:rtype: pyplanet.contrib.command.command.Command
		"""
		self.parser.add_param(
			name=name, nargs=nargs, type=type, default=default, required=required, help=help, dest=dest
		)
		return self

	async def handle(self, instance, player, argv):
		"""
		Handle command parsing and execution.

		:param player: Player object.
		:param argv: Arguments in array
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		"""
		# Check permissions.
		if not await self.has_permission(instance, player):
			await instance.chat(
				'$z$sYou are not authorized to use this command!',
				player.login
			)
			return

		# Strip off the namespace and command.
		paramv = self.get_params(argv)

		# Parse, validate and show errors if any.
		self.parser.parse(paramv)
		if not self.parser.is_valid():
			await instance.gbx.multicall(
				instance.chat('$z$sCommand operation got invalid arguments: {}'.format(', '.join(self.parser.errors)), player),
				instance.chat('$z$s >> {}'.format(self.usage_text), player),
			)
			return

		# We are through. Call our target!
		if iscoroutinefunction(self.target):
			return await self.target(player=player, data=self.parser.data, raw=argv, command=self)
		return self.target(player=player, data=self.parser.data, raw=argv, command=self)

	async def has_permission(self, instance, player):
		"""
		Checks whether the provided player has the permission to execute this command.
		:param instance: Controller Instance
		:type instance: pyplanet.core.instance.Instance
		:param player: Player requesting execution of this command.
		:type player: pyplanet.apps.core.maniaplanet.models.player.Player
		:return: Whether provided player has permission to execute this command.
		"""
		player_has_permission = True
		if self.perms and len(self.perms) > 0:
			# All the given perms need to be matching!
			is_allowed = await asyncio.gather(*[
				instance.permission_manager.has_permission(player, perm) for perm in self.perms
			])
			if not all(allowed is True for allowed in is_allowed):
				player_has_permission = False

		return player_has_permission

	@property
	def usage_text(self):
		"""
		The usage text line for the command.
		"""
		text = 'Usage: /{}{}{}'.format(
			'/' if self.admin else '',
			self.namespace if self.namespace else '',
			self.command
		)

		for param in self.parser.params:
			text += ' {}{}:{}{}'.format(
				'[' if not param['required'] else '',
				param['name'],
				getattr(param['type'], '__name__', 'any'),
				']' if not param['required'] else '',
			)

		return text

	@property
	def params_text(self):
		text = ''

		param_index = 0
		for param in self.parser.params:
			if param_index > 0:
				text += '\n'

			text += '{}{}:{}{}{}'.format(
				'[' if not param['required'] else '',
				param['name'],
				getattr(param['type'], '__name__', 'any'),
				']' if not param['required'] else '',
				' = {}'.format(param['help']) if param['help'] else ''
			)

			param_index += 1

		return text

	@property
	def perms_text(self):
		text = ''

		if self.perms and len(self.perms) > 0:
			perm_index = 0
			for permission in self.perms:
				if perm_index > 0:
					text += '\n'

				text += '{}'.format(permission)

				perm_index += 1

		return text

	def __str__(self):
		# Make sure namespace is always an array if provided.
		if self.namespace and not isinstance(self.namespace, (list, tuple)):
			self.namespace = [self.namespace]

		return '/{}{}{}'.format(
			'/' if self.admin else '',
			'|'.join(self.namespace) if self.namespace and isinstance(self.namespace, (list, tuple)) else self.command,
			' ' + self.command if self.namespace else '',
		)



import asyncio_extras
import os

import importlib
from async_generator import yield_

from pyplanet.conf import settings
from pyplanet.core.storage import StorageDriver, StorageInterface


class Storage(StorageInterface):
	"""
	The storage component manager is managing the storage access trough drivers that can be customized.
	
	.. warning::
	
		Some drivers are work in progress!

	"""
	MAP_FOLDER = 'UserData/Maps'
	MATCHSETTINGS_FOLDER = 'UserData/Maps/MatchSettings'

	def __init__(self, instance, driver: StorageDriver, config):
		"""
		Initiate storage manager.
		
		:param instance: Instance of the controller.
		:param driver: Driver instance, must be init already!
		:param config: Storage configuration (including driver + driver config).
		:type instance: pyplanet.core.instance.Instance
		:type driver: pyplanet.core.storage.interface.StorageDriver
		:type config: dict
		"""
		self._instance = instance
		self._driver = driver
		self._config = config
		self._game = None

		# Create temp folders for driver.
		self._tmp_root = os.path.join(settings.TMP_PATH, self._instance.process_name)
		self._tmp_driver = os.path.join(self._tmp_root, )

	@classmethod
	def create_from_settings(cls, instance, storage_config):
		driver_path, _, driver_cls_name = storage_config['DRIVER'].rpartition('.')
		driver_options = storage_config['OPTIONS'] if 'OPTIONS' in storage_config else dict()
		driver_cls = getattr(importlib.import_module(driver_path), driver_cls_name)
		driver = driver_cls(instance, driver_options)
		return cls(instance, driver, storage_config)

	async def initialize(self):
		self._game = self._instance.game
		self._driver.map_dir = self._game.server_map_dir
		self._driver.skin_dir = self._game.server_skin_dir
		self._driver.data_dir = self._game.server_data_dir
		self._driver.base_dir = self._game.server_data_dir[:len(self._game.server_data_dir)-9]

	@property
	def driver(self):
		"""
		Get the raw driver. Be careful with this!
		
		:return: Driver Instance
		:rtype: pyplanet.core.storage.interface.StorageDriver
		"""
		return self._driver

	@asyncio_extras.async_contextmanager
	async def open(self, file: str, mode: str = 'rb', **kwargs):
		"""
		Open a file on the server. Use relative path to the dedicated root. Use the other open methods to relative
		from another base path.
		
		:param file: Filename/path, relative to the dedicated root path.
		:param mode: Mode to open, see the python `open` manual for supported modes.
		:return: File handler.
		"""
		context = self._driver.open(file, mode, **kwargs)
		await yield_(await context.__aenter__())
		await context.__aexit__(None, None, None)

	@asyncio_extras.async_contextmanager
	async def open_match_settings(self, file: str, mode: str = 'r', **kwargs):
		"""
		Open a file on the server. Relative to the MatchSettings folder (UserData/Maps/MatchSettings).
		
		:param file: Filename/path, relative to the dedicated matchsettings folder.
		:param mode: Mode to open, see the python `open` manual for supported modes.
		:return: File handler.
		"""
		context = self._driver.open('{}/{}'.format(self.MATCHSETTINGS_FOLDER, file), mode, **kwargs)
		await yield_(await context.__aenter__())
		await context.__aexit__(None, None, None)

	@asyncio_extras.async_contextmanager
	async def open_map(self, file: str, mode: str = 'rb', **kwargs):
		"""
		Open a file on the server. Relative to the Maps folder (UserData/Maps).
		
		:param file: Filename/path, relative to the dedicated maps folder.
		:param mode: Mode to open, see the python `open` manual for supported modes.
		:return: File handler.
		"""
		context = self._driver.open('{}/{}'.format(self.MAP_FOLDER, file), mode, **kwargs)
		await yield_(await context.__aenter__())
		await context.__aexit__(None, None, None)

	async def remove_map(self, file: str):
		"""
		Remove a map file with filename given.
		
		:param file: Filename, relative to Maps folder.
		"""
		await self._driver.remove('{}/{}'.format(self.MAP_FOLDER, file))



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
#
# PyPlanet documentation build configuration file, created by
# sphinx-quickstart on Thu Mar 16 11:38:18 2017.
#
# This file is execfile()d with the current directory set to its
# containing dir.
#
# Note that not all possible configuration values are present in this
# autogenerated file.
#
# All configuration values have a default; values that are commented out
# serve to show the default.

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys
import sphinx_rtd_theme

src_folder = os.path.abspath(os.path.join(os.path.dirname(__name__), '..', '..'))
sys.path.insert(0, src_folder)

# Import now, we won't have issues with circular imports
import pyplanet
import pyplanet.core.instance
import pyplanet.apps

# -- General configuration ------------------------------------------------

# If your documentation needs a minimal Sphinx version, state it here.
#
# needs_sphinx = '1.0'

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
	'sphinx.ext.autodoc',
	'sphinx.ext.doctest',
	'sphinx.ext.todo',
	'sphinx.ext.coverage',
	'sphinx.ext.imgmath',
	'sphinx.ext.viewcode',
	'sphinx.ext.githubpages'
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ['_templates']

# The suffix(es) of source filenames.
# You can specify multiple suffix as a list of string:
source_suffix = ['.rst', '.md']

# The master toctree document.
master_doc = 'index'

# General information about the project.
project = 'PyPlanet'
copyright = '2017, PyPlanet (Tom Valk)'
author = 'Tom Valk'

# The version info for the project you're documenting, acts as replacement for
# |version| and |release|, also used in various other places throughout the
# built documents.
#
# The short X.Y version.
version = '0.9.13-rc1'
# The full version, including alpha/beta/rc tags.
release = '0.9.13-rc1'

# The language for content autogenerated by Sphinx. Refer to documentation
# for a list of supported languages.
#
# This is also used if you do content translation via gettext catalogs.
# Usually you set "language" from the command line for these cases.
language = None

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This patterns also effect to html_static_path and html_extra_path
exclude_patterns = []

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = 'sphinx'

# If true, `todo` and `todoList` produce output, else they produce nothing.
todo_include_todos = True
todo_link_only = True

# -- Options for HTML output ----------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = 'sphinx_rtd_theme'

# Theme options are theme-specific and customize the look and feel of a theme
# further.  For a list of options available for each theme, see the
# documentation.
#
# html_theme_options = {}
html_theme_path = [sphinx_rtd_theme.get_html_theme_path()]

# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ['_static']

html_context = {
	'display_github': True,
	'github_user': 'PyPlanet',
	'github_repo': 'pyplanet',
	'github_version': 'master',
	'conf_py_path': '/docs/source/'
}

# -- Options for HTMLHelp output ------------------------------------------

# Output file base name for HTML help builder.
htmlhelp_basename = 'PyPlanetdoc'

# -- Options for LaTeX output ---------------------------------------------

latex_elements = {
	# The paper size ('letterpaper' or 'a4paper').
	#
	# 'papersize': 'letterpaper',

	# The font size ('10pt', '11pt' or '12pt').
	#
	# 'pointsize': '10pt',

	# Additional stuff for the LaTeX preamble.
	#
	# 'preamble': '',

	# Latex figure (float) alignment
	#
	# 'figure_align': 'htbp',
}

# Grouping the document tree into LaTeX files. List of tuples
# (source start file, target name, title,
#  author, documentclass [howto, manual, or own class]).
latex_documents = [
	(master_doc, 'PyPlanet.tex', 'PyPlanet Documentation',
	 'PyPlanet (Tom Valk)', 'manual'),
]

# -- Options for manual page output ---------------------------------------

# One entry per manual page. List of tuples
# (source start file, name, description, authors, manual section).
man_pages = [
	(master_doc, 'pyplanet', 'PyPlanet Documentation',
	 [author], 1)
]

# -- Options for Texinfo output -------------------------------------------

# Grouping the document tree into Texinfo files. List of tuples
# (source start file, target name, title, author,
#  dir menu entry, description, category)
texinfo_documents = [
	(master_doc, 'PyPlanet', 'PyPlanet Documentation',
	 author, 'PyPlanet', 'One line description of project.',
	 'Miscellaneous'),
]

# -- Options for Epub output ----------------------------------------------

# Bibliographic Dublin Core info.
epub_title = project
epub_author = author
epub_publisher = author
epub_copyright = copyright

# The unique identifier of the text. This can be a ISBN number
# or the project homepage.
#
# epub_identifier = ''

# A unique identification for the text.
#
# epub_uid = ''

# A list of files that should not be packed into the epub file.
epub_exclude_files = ['search.html']



import datetime

from pyplanet.apps.core.maniaplanet.models import Player, Map
from pyplanet.apps.contrib.jukebox.views import FolderListView, FolderMapListView
from pyplanet.contrib.setting import Setting

from .models import MapFolder as Folders, MapInFolder


class FolderManager:
	def __init__(self, app):
		self.app = app

		# Initiate global folders.
		self.auto_folders = list()
		self.public_folders = list()

	async def on_start(self):
		"""
		Called after startup of PyPlanet so it can investigate the automatic folders.

		:return:
		"""
		await self.update_folders()

	async def update_folders(self, *args, **kwargs):
		"""
		Update the auto-folders.

		:return:
		"""
		# Prepare variables.
		self.auto_folders = list()

		days_filter = await self.app.setting_newest_days_range.get_value()
		self.auto_folders.append({
			'id': 'newest', 'name': 'Newest maps (last {} days)'.format(days_filter), 'owner': 'PyPlanet', 'type': 'auto'
		})

		if 'local_records' in self.app.instance.apps.apps:
			self.auto_folders.append({'id': 'local_none', 'name': 'Map record: none', 'owner': 'PyPlanet', 'type': 'auto'})
			self.auto_folders.append({'id': 'length_shorter_30s', 'name': 'Map record: below 30 seconds', 'owner': 'PyPlanet', 'type': 'auto'})
			self.auto_folders.append({'id': 'length_longer_60s', 'name': 'Map record: above 60 seconds', 'owner': 'PyPlanet', 'type': 'auto'})

		if 'karma' in self.app.instance.apps.apps:
			self.auto_folders.append({'id': 'karma_none', 'name': 'Map karma: no votes', 'owner': 'PyPlanet', 'type': 'auto'})
			self.auto_folders.append({'id': 'karma_negative', 'name': 'Map karma: negative', 'owner': 'PyPlanet', 'type': 'auto'})
			self.auto_folders.append({'id': 'karma_undecided', 'name': 'Map karma: undecided', 'owner': 'PyPlanet', 'type': 'auto'})
			self.auto_folders.append({'id': 'karma_positive', 'name': 'Map karma: positive', 'owner': 'PyPlanet', 'type': 'auto'})

	async def get_folders(self, player):
		"""
		Get the folders, as combined object, for the specific player.

		:param player: Player instance
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return:
		"""
		# Fetch the public or private folders of the users.
		raw_list = await Folders.objects.execute(
			Folders.select(Folders, Player)
				.join(Player)
				.where((Player.login == player.login) | (Folders.public == True))
				.order_by(Folders.public.desc())
		)

		# Convert to the wanted objects.
		folder_list = list()
		folder_list.extend(self.auto_folders.copy())

		for folder in raw_list:
			folder_id = 'database_{}'.format(folder.get_id())
			folder_list.append(
				{
					'id': folder_id, 'name': folder.name, 'owner_login': folder.player.login,
					'owner': folder.player.nickname, 'type': 'public' if folder.public else 'private'
				}
			)

		return folder_list

	async def get_private_folders(self, player):
		"""
		Get the private folders of the given player.

		:param player: Player instance.
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return:
		"""
		return await Folders.objects.execute(
			Folders.select(Folders, Player)
				.join(Player)
				.where((Player.login == player.login) & (Folders.public == False))
		)

	async def get_writable_folders(self, player):
		"""
		Get writable folders for the given player.

		:param player: Player instance
		:type player: pyplanet.apps.core.maniaplanet.models.Player
		:return: List with writable folders.
		"""
		if player.level >= player.LEVEL_ADMIN:
			return await Folders.objects.execute(
				Folders.select(Folders, Player)
					.join(Player)
					.where((Player.login == player.login) | (Folders.public == True))
			)

		return await Folders.objects.execute(
			Folders.select(Folders, Player)
				.join(Player)
				.where((Player.login == player.login))
		)

	async def create_folder(self, **kwargs):
		"""
		Create folder, based on the paramters given.

		:param kwargs:
		:return: Folder instance.
		"""
		folder = Folders(**kwargs)
		await folder.save()
		return folder

	async def display_folder_list(self, player):
		"""
		Create folder listview, display and return instance.

		:param player:
		:return:
		"""
		view = FolderListView(self, player)
		await view.display(player=player)
		return view

	async def add_map_to_folder(self, folder_id, map_id):
		"""
		Adds the provided map to the provided folder.

		:param folder_id: folder identifier to add the map to
		:param map_id: map identifier to add to the folder
		:return: whether the map was successfully added to the folder
		"""
		existing = await MapInFolder.execute(
			MapInFolder.select(MapInFolder)
				.where((MapInFolder.folder_id == folder_id) & (MapInFolder.map_id == map_id))
		)

		if existing:
			return False

		# Add map to folder.
		map_in_folder = MapInFolder(
			map_id=map_id,
			folder_id=folder_id
		)
		await map_in_folder.save()

		return True

	async def remove_map_from_folder(self, folder_id, map_id):
		"""
		Removes the provided map from the provided folder.

		:param folder_id: folder identifier to remove map from
		:param map_id: map identifier to remove from folder
		"""
		await MapInFolder.execute(
			MapInFolder.delete()
				.where((MapInFolder.map_id == map_id) & (MapInFolder.folder_id == folder_id))
		)

	async def get_folders_containing_map(self, map_id):
		"""
		Returns the folders that contain the provided map.

		:param map_id: map identifier to check folders for
		:return: list of Folder instances (None if no found)
		"""
		map_in_folders = await MapInFolder.execute(
			MapInFolder.select(MapInFolder)
				.where(MapInFolder.map_id == map_id)
		)

		if map_in_folders is None or len(map_in_folders) == 0:
			return None

		folder_ids = [mf.folder_id for mf in map_in_folders]

		return await Folders.execute(
			Folders.select(Folders)
				.where(Folders.id << folder_ids)
		)

	async def get_folder_code_contents(self, folder_code):
		folder = folder_code

		map_list = []
		fields = []
		folder_instance = None

		if folder['id'] == 'newest':
			days_ago = await self.app.setting_newest_days_range.get_value()
			filter_from = datetime.datetime.now() - datetime.timedelta(days=days_ago)
			map_list = [m for m in self.app.instance.map_manager.maps if m.created_at > filter_from]
		elif folder['id'] == 'local_none':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'local') and m.local['record_count'] == 0]
		elif folder['id'] == 'length_shorter_30s':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'local') and m.local['first_record'] is not None and m.local['first_record'].score < 30000]
		elif folder['id'] == 'length_longer_60s':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'local') and m.local['first_record'] is not None and m.local['first_record'].score > 60000]
		elif folder['id'] == 'karma_none':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'karma') and m.karma['vote_count'] == 0]
		elif folder['id'] == 'karma_negative':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'karma') and m.karma['map_karma'] < 0]
		elif folder['id'] == 'karma_undecided':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'karma') and m.karma['map_karma'] == 0]
		elif folder['id'] == 'karma_positive':
			map_list = [m for m in self.app.instance.map_manager.maps if hasattr(m, 'karma') and m.karma['map_karma'] > 0]
		elif folder['id'].startswith('database_'):
			# Get the real folder model instance.
			folder_instance = await Folders.get(id=int(folder['id'].replace('database_', '')))

			# Personal folder from database
			maps_in_folder = [m.map.uid for m in await MapInFolder.objects.execute(
				MapInFolder.select(MapInFolder, Map)
					.join(Map)
					.where(MapInFolder.folder == int(folder['id'].replace('database_', '')))
			)]

			map_list = [m for m in self.app.instance.map_manager.maps if m.uid in maps_in_folder]

		if folder['id'].startswith('length_') or folder['id'].startswith('database_'):
			fields.append({
				'name': 'Local Record',
				'index': 'local_record',
				'sorting': True,
				'searching': False,
				'width': 40,
			})

		if folder['id'].startswith('karma_') or folder['id'].startswith('database_'):
			fields.append({
				'name': 'Karma',
				'index': 'karma',
				'sorting': True,
				'searching': False,
				'width': 20,
			})

		return fields, map_list, folder, folder_instance

	async def display_folder(self, player, folder_code):
		# Initiate folder contents list view.
		view = FolderMapListView(self, folder_code, player)
		await view.display(player=player)



import inspect
import importlib
import logging
import os

from pyplanet.core.exceptions import ImproperlyConfigured, InvalidAppModule


class _AppContext:
	"""
	The app context holds instances of core/contrib components that must be managed on a per app base. Such as the UI
	registration and distribution.
	"""
	def __init__(self, app):
		"""
		Initiate the App Context. Used by several core and contribs to have it's own manager instance per app.
		You should always use the managers of your local app at first!

		:param app: App Config instance.
		:type app: pyplanet.apps.config.AppConfig
		"""
		self.ui = app.instance.ui_manager.create_app_manager(app)
		"""
		UI Component. See :doc:`UI Classes </api/core_ui>`.
		"""

		self.setting = app.instance.setting_manager.create_app_manager(app)
		"""
		Setting Contrib Component. See :doc:`Setting Classes </api/contrib_setting>`.
		"""

		self.signals = app.instance.signals.create_app_manager(app)
		"""
		Signal manager. See :doc:`Signal Manager </api/core_events>`.
		"""

	async def on_destroy(self):
		await self.ui.on_destroy()
		await self.signals.on_destroy()


class AppState:
	UNLOADED = 0
	LOADED = 1


class AppConfig:
	"""
	This class is the base class for the Applications metadata class. The class holds information and hooks
	that will be executed after initiation for example.

	.. code-block:: python

		class MyApp(AppConfig):

			async def on_start(self):
				print('we are staring!!')

	"""

	name = None
	label = None
	human_name = None
	path = None

	app_dependencies = None
	"""
	You can provide a list of dependencies to other apps (each entry needs to be a string of the app label!)
	"""

	mode_dependencies = None
	"""
	You can provide a list of gamemodes that are required to activate the app. Gamemodes needs to be declared as
	script names.
	You can override this behaviour by defining the following method in your config class

	.. code-block :: python

		def is_mode_supported(self, mode):
			return mode.lower().startswith('TimeAttack')

	"""

	game_dependencies = None
	"""
	You can provide a list of game dependencies that needs to meet when the app is started. For example you can provide:

	.. code-block :: python

		game_dependencies = ['trackmania']

	You can override this behaviour by defining the following method in your config class

	.. code-block :: python

		def is_game_supported(self, game):
			return game != 'questmania'

	"""

	def __init__(self, app_name, app_module, instance):
		"""
		Init app config.

		:param app_name: App Name (from module path).
		:param app_module: App Module.
		:param instance: Instance of controller
		:type app_name: str
		:type app_module: str
		:type instance: pyplanet.core.instance.Instance
		"""
		# The full python module path. The postfix `*.app` is always the same!
		# Example: pyplanet.contrib.apps.games.trackmania.app
		self.name = app_name

		# The apps root module.
		# Example: pyplanet.contrib.apps.games.trackmania
		self.module = app_module

		# The apps registry will be injected into the app config.
		self.apps = None

		# Make sure we give the core attribute the default value of false. This indicates if it's an internally
		# module.
		self.core = getattr(self, 'core', False)

		# The label can be given by the module, or automatically determinated on the last component.
		if not hasattr(self, 'label') or getattr(self, 'label', None) is None:
			self.label = app_name.rpartition('.')[2]

			# If the module is a core contrib module, we give the label a prefix (contrib.app).
			if self.core is True:
				self.label = 'core.{}'.format(self.label)

		# Human-readable name for the application eg. `MyApp`.
		if not hasattr(self, 'human_name') or getattr(self, 'human_name', None) is None:
			self.human_name = self.label.title()

		# Filesystem path to the application directory eg.
		if not hasattr(self, 'path') or getattr(self, 'path') is None:
			self.path = self._path_from_module(app_module)

		# The instance and related app context managers.
		self.instance = instance
		self.context = _AppContext(self)

		# State of app.
		self.state = AppState.UNLOADED

	def __repr__(self):
		return '<%s: %s>' % (self.__class__.__name__, self.label)

	###################################################################################################
	# Lifecycle Methods
	###################################################################################################

	async def on_init(self):
		"""
		The on_init will be called before all apps are started (just before the on_ready). This will allow the app
		to register things like commands, permissions and other things that are important and don't require other
		apps to be ready.
		"""

	async def on_start(self):
		"""
		The on_start call is being called after all apps has been started successfully. You should register any stuff
		that is related to any other apps and signals like your `self` context for signals if they are classmethods.
		"""
		# Deprecated: Fix the deprecated method
		if hasattr(self, 'on_ready'):
			logging.warning('on_ready is deprecated, use on_start instead! app: {}'.format(self.label))
			await self.on_ready()

	async def on_stop(self):
		"""
		The on_stop will be called before stopping the app.
		"""
		pass

	async def on_destroy(self):
		"""
		On destroy is being called when unloading the app from the memory.
		"""
		await self.context.on_destroy()

	###################################################################################################

	def is_mode_supported(self, mode):
		if self.mode_dependencies:
			for mode_requirement in self.mode_dependencies:
				if mode_requirement.lower() in mode.lower():
					return True
			return False
		return True

	def is_game_supported(self, game):
		if self.game_dependencies:
			return game in self.game_dependencies
		return True

	def _path_from_module(self, module):
		"""Attempt to determine app's filesystem path from its module."""
		paths = list(getattr(module, '__path__', []))

		if len(paths) != 1:
			filename = getattr(module, '__file__', None)
			if filename is not None:
				paths = [os.path.dirname(filename)]
			else:
				# Can be bugged for unknown reasons.
				paths = list(set(paths))

		if len(paths) > 1:
			raise ImproperlyConfigured(
				'The app module {} has multiple filesystem locations {}; '
				'you must configure this app with an AppConfig subclass '
				'with a \'path\' class attribute.'.format(module, paths))

		elif not paths:
			raise ImproperlyConfigured(
				'The app module {} has no filesystem location, '
				'you must configure this app with an AppConfig subclass '
				'with a \'path\' class attribute.'.format(module))

		return paths[0]

	@staticmethod
	def import_app(entry, instance):
		# Import the module, we need to strip down the path into namespace, file and class.
		module_path, app_glue, cls_name = entry.rpartition('.')

		# The app name (full module path to module, not the app!)
		app_name = module_path.rpartition('.')[0]

		if not module_path:
			raise ImproperlyConfigured('Module for your app {} can\'t be found!'.format(entry))

		# The new style definitions works a bit different. We just import the module, search for the first class that is
		# a subclass of AppConfig.
		if cls_name.islower():
			module_path += '.' + cls_name
			app_name = module_path
			cls_name = None

		# Try to load the app module, containing the class.
		try:
			module = importlib.import_module(module_path)

			# If we have a new-style module, check for the first AppConfig extending class in our module.
			# See #109.
			if cls_name is None:
				for name, obj in inspect.getmembers(module):
					if inspect.isclass(obj) and issubclass(obj, AppConfig) and obj.__name__ != 'AppConfig':
						cls_name = obj.__name__
						break

			module = getattr(module, cls_name)
		except TypeError as e:
			raise ImproperlyConfigured(
				'Can\'t load the app {}. Can\'t find the app config! '
				'Check your apps.py and if your app is installed correctly'.format(entry)
			) from e
		except ImportError:
			raise ImproperlyConfigured(
				'Can\'t load the app {}. Can\'t find the app config!'.format(entry)
			)
		except AttributeError as e:
			raise ImproperlyConfigured(
				'Can\'t load the app {}. Can\'t load the app class!'.format(entry)
			) from e

		# Last check if subclass of appconfig.
		if not issubclass(module, AppConfig):
			raise InvalidAppModule('Your required app {} couldn\'t be loaded!'.format(entry))

		# Ensure app_name points to a valid module.
		try:
			app_module = importlib.import_module(app_name)
		except ImportError:
			raise ImproperlyConfigured(
				'Can\'t import {}. Check that \'{}.{}.name\' is correct.'.format(
					app_name, module_path, cls_name
				)
			)

		return module(app_name, app_module, instance)



from watchdog.events import FileSystemEventHandler, FileModifiedEvent


class LiveReload(FileSystemEventHandler):
	def __init__(self, pool):
		"""
		:param pool:
		:type pool: pyplanet.god.pool.EnvironmentPool
		"""
		self.pool = pool

	def on_any_event(self, event):
		if isinstance(event, FileModifiedEvent):
			if event.src_path.endswith('.py'):
				self.pool.restart()




from .dispatcher import Signal
from .manager import SignalManager, public_signal, public_callback
from .callback import Callback, handle_generic

__all__ = [
	'Signal',

	'SignalManager',
	'public_signal',
	'public_callback',

	'Callback',
]
